import os
import time

import sys
import traceback

from PyQt5 import QtCore
from contextlib import contextmanager
from functools import wraps

from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QApplication

from mdt.__version__ import __version__
from mdt.lib.log_handlers import LogListenerInterface

__author__ = 'Robbert Harms'
__date__ = "2015-08-20"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class QtManager(object):

    windows = []

    @staticmethod
    def get_qt_application_instance():
        q_app = QApplication.instance()
        if q_app is None:
            q_app = QApplication([])
            q_app.lastWindowClosed.connect(QtManager.empty_windows_list)
        return q_app

    @staticmethod
    def exec_():
        if QtManager.windows:
            QtManager.get_qt_application_instance().exec_()

    @staticmethod
    def add_window(window):
        QtManager.windows.append(window)

    @staticmethod
    def empty_windows_list():
        QtManager.windows = []


def center_window(window):
    """Center the given window on the screen.

    Args:
        q_app (QApplication): for desktop information
        window (QMainWindow): the window to center
    """
    q_app = QApplication.instance()

    frame_gm = window.frameGeometry()
    screen = q_app.desktop().screenNumber(q_app.desktop().cursor().pos())
    center_point = q_app.desktop().screenGeometry(screen).center()
    frame_gm.moveCenter(center_point)
    window.move(frame_gm.topLeft())


def function_message_decorator(header, footer):
    """This creates and returns a decorator that prints a header and footer before executing the function.

    Args:
        header (str): the header text, we will add extra decoration to it
        footer (str): the footer text, we will add extra decoration to it

    Returns:
        decorator function
    """
    def _called_decorator(dec_func):

        @wraps(dec_func)
        def _decorator(*args, **kwargs):
            print('')
            print(header)
            print('-'*20)

            response = dec_func(*args, **kwargs)

            print('-'*20)
            print(footer)

            return response
        return _decorator

    return _called_decorator


@contextmanager
def blocked_signals(*widgets):
    """Small context in which the signals of the given widget are blocked.

    Args:
        widgets (QWidget): one or more widgets
    """
    def apply_block(bool_val):
        for w in widgets:
            w.blockSignals(bool_val)

    apply_block(True)
    yield
    apply_block(False)


def print_welcome_message():
    """Prints a small welcome message for after the GUI has loaded.

    This prints to stdout. We expect the GUI to catch the stdout events and redirect them to the GUI.
    """
    from mdt import VERSION
    print('Welcome to MDT version {}.'.format(VERSION))
    print('')
    print('This area is reserved for log output.')
    print('-------------------------------------')


class ForwardingListener(LogListenerInterface):

    def __init__(self, queue):
        """Forwards all incoming messages to the given _logging_update_queue.

        Instances of this class can be used as a log listener to the MDT LogDispatchHandler and as a
        sys.stdout replacement.

        Args:
            queue (Queue): the _logging_update_queue to forward the messages to
        """
        self._queue = queue

    def emit(self, record, formatted_message):
        self._queue.put(formatted_message + "\n")

    def write(self, string):
        self._queue.put(string)


image_files_filters = ['All files (*)',
                       'Nifti (*.nii *.nii.gz)',
                       'IMG, HDR (*.img)']
protocol_files_filters = ['MDT protocol (*.prtcl)',
                          'Text files (*.txt)',
                          'All files (*)']


class UpdateDescriptor(object):

    def __init__(self, attribute_name):
        """Descriptor that will emit a state_updated_signal at each update.

        This accesses from the instance the attribute name prepended with an underscore (_).
        """
        self._attribute_name = attribute_name

    def __get__(self, instance, owner):
        return getattr(instance, '_' + self._attribute_name)

    def __set__(self, instance, value):
        setattr(instance, '_' + self._attribute_name, value)
        instance.state_updated_signal.emit(self._attribute_name)


class MessageReceiver(QObject):

    text_message_signal = pyqtSignal(str)
    finished = pyqtSignal()

    def __init__(self, queue, *args, **kwargs):
        """A QObject (to be run in a QThread) which sits waiting for data to come through a Queue.Queue().

        It blocks until data is available, and one it has got something from the _logging_update_queue, it sends
        it to the "MainThread" by emitting a Qt Signal.

        Attributes:
            is_running (boolean): set to False to stop the receiver.
        """
        super().__init__(*args, **kwargs)
        self.queue = queue
        self.is_running = True

    def run(self):
        while self.is_running:
            if not self.queue.empty():
                self.text_message_signal.emit(self.queue.get())
            time.sleep(0.001)
        self.finished.emit()


class MainTab(object):

    def tab_opened(self):
        """Called when this tab is selected by the user."""


class TimedUpdate(QTimer):

    def __init__(self, update_cb):
        """Creates a timer that can delay running a given callback function.

        Every time the user adds a delayed callback the timer gets reset to the new value and we will
        wait that new value until calling the callback with the last data given.

        Args:
            update_cb (function): the function we would like to run after a timer has run out
        """
        super().__init__()
        self._cb_values = []
        self._update_cb = update_cb
        self.timeout.connect(self._call_update_cb)
        self.timeout.connect(self.stop)

    def add_delayed_callback(self, delay, *cb_values):
        """Pushes a new delay to calling the callback function.

        Args:
            delay (int): the time in ms to wait
            cb_values (*list): the list of values to use as arguments to the callback function. Leave empty to disable.
        """
        self._cb_values = cb_values
        self.start(delay)

    def _call_update_cb(self):
        if self._cb_values:
            self._update_cb(*self._cb_values)
        else:
            self._update_cb()


def get_script_file_header_text(info_fields=None):
    """Get the header file text to use when outputting script files matching GUI actions.

    Args:
        info_fields (dict): fields holding information. Common fields are 'Purpose' and 'Example usage'.
            These fields are appended to the header text.

    Returns:
        str: the header (comment) text for the automatically generated script.
    """
    header_text = '# This script was automatically generated by the MDT GUI. \n'
    header_text += '# \n'
    header_text += '# {}: {}\n'.format('Generated on', time.strftime("%Y-%m-%d %H:%M:%S"))
    header_text += '# {}: {}\n'.format('MDT version', __version__)

    for key, value in info_fields.items():
        header_text += '# {key}: {value} \n'.format(key=key, value=value)

    return header_text[:-1]


def split_long_path_elements(original_path, max_single_element_length=25):
    """Split long path elements into smaller ones using spaces

    Args:
        original_path (str): the path you want to split
        max_single_element_length (int): the maximum length allowed per path component (folders and filename).

    Returns:
        str: the same path but with spaces in long path elements. The result will no longer be a valid path.
    """

    def split(p):
        listing = []

        def _split(el):
            if el:
                head, tail = os.path.split(el)
                if not tail:
                    listing.append(head)
                else:
                    _split(head)
                    listing.append(tail)

        _split(p)
        return listing

    elements = list(split(original_path))
    new_elements = []

    for el in elements:
        if len(el) > max_single_element_length:
            item = ''
            for i in range(0, len(el), max_single_element_length):
                item += el[i:i + max_single_element_length] + ' '
            item = item[:-1]
            new_elements.append(item)
        else:
            new_elements.append(el)

    return os.path.join(*new_elements)


def enable_pyqt_exception_hook():
    """Enable the PyQt exception handling hook for PyQt versions larger than 5.5.

    If this is not enabled, exceptions will be handled silently and will not be printed to the user. This
    makes it harder to solve the issue.
    """
    if QtCore.QT_VERSION >= 0x50501:
        old_stdout = sys.stdout
        old_stderr = sys.stderr

        def excepthook(type_, value, traceback_):
            sys.stdout = old_stdout
            sys.stderr = old_stderr
            traceback.print_exception(type_, value, traceback_)
            QtCore.qFatal('')
        sys.excepthook = excepthook



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Generate the bval and bvec file from a protocol file."""
import argparse
import os
import mdt
from argcomplete.completers import FilesCompleter
import textwrap
import mdt.protocols
from mdt.lib.shell_utils import BasicShellApplication
from mdt.protocols import write_bvec_bval

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CreateBvecBval(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        examples = textwrap.dedent('''
            mdt-create-bvec-bval my_protocol.prtcl
            mdt-create-bvec-bval my_protocol.prtcl bvec_name.bvec bval_name.bval
        ''')
        epilog = self._format_examples(doc_parser, examples)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('protocol', help='the protocol file').completer = FilesCompleter()
        parser.add_argument('bvec', help="the output bvec file", nargs='?', default=None).completer = FilesCompleter()
        parser.add_argument('bval', help="the output bvec file", nargs='?', default=None).completer = FilesCompleter()

        return parser

    def run(self, args, extra_args):
        protocol_base = os.path.join(os.path.dirname(os.path.realpath(args.protocol)),
                                     os.path.splitext(os.path.basename(args.protocol))[0])

        if args.bvec:
            bvec = os.path.realpath(args.bvec)
        else:
            bvec = protocol_base + '.bvec'

        if args.bval:
            bval = os.path.realpath(args.bval)
        else:
            bval = protocol_base + '.bval'

        write_bvec_bval(mdt.load_protocol(os.path.realpath(args.protocol)), bvec, bval)


def get_doc_arg_parser():
    return CreateBvecBval().get_documentation_arg_parser()


if __name__ == '__main__':
    CreateBvecBval().start()



from mdt import CompositeModelTemplate
import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BallStick_r1(CompositeModelTemplate):

    description = 'The default Ball & Stick model'
    model_expression = '''
        S0 * ( (Weight(w_ball) * Ball) +
               (Weight(w_stick0) * Stick(Stick0)) )
    '''
    fixes = {'Ball.d': 3.0e-9,
             'Stick0.d': 1.7e-9}

    extra_optimization_maps = [
        lambda results: {'FS': results['w_stick0.w']},
        lambda results: {'FS.std': results['w_stick0.w.std']}
    ]
    extra_sampling_maps = [
        lambda samples: {'FS': np.mean(samples['w_stick0.w'], axis=1),
                         'FS.std': np.std(samples['w_stick0.w'], axis=1)}
    ]


class BallStick_r2(CompositeModelTemplate):

    description = 'The Ball & 2x Stick model'
    model_expression = '''
        S0 * ( (Weight(w_ball) * Ball) +
               (Weight(w_stick0) * Stick(Stick0)) +
               (Weight(w_stick1) * Stick(Stick1)) )
    '''
    fixes = {'Ball.d': 3.0e-9,
             'Stick0.d': 1.7e-9,
             'Stick1.d': 1.7e-9}

    sort_maps = [('w_stick0', 'w_stick1'), ('Stick0', 'Stick1')]

    extra_optimization_maps = [
        lambda results: {'FS': 1 - results['w_ball.w']},
        lambda results: {'FS.std': results['w_ball.w.std']}
    ]
    extra_sampling_maps = [
        lambda samples: {
            'FS': np.mean(samples['w_stick0.w'] + samples['w_stick1.w'], axis=1),
            'FS.std': np.std(samples['w_stick0.w'] + samples['w_stick1.w'], axis=1)}
    ]

    extra_prior = 'return w_stick1.w < w_stick0.w;'


class BallStick_r3(CompositeModelTemplate):

    description = 'The Ball & 3x Stick model'
    model_expression = '''
            S0 * ( (Weight(w_ball) * Ball) +
                   (Weight(w_stick0) * Stick(Stick0)) +
                   (Weight(w_stick1) * Stick(Stick1)) +
                   (Weight(w_stick2) * Stick(Stick2)) )
        '''
    fixes = {'Ball.d': 3.0e-9,
             'Stick0.d': 1.7e-9,
             'Stick1.d': 1.7e-9,
             'Stick2.d': 1.7e-9}
    inits = {'w_stick2.w': 0}

    sort_maps = [('w_stick0', 'w_stick1', 'w_stick2'), ('Stick0', 'Stick1', 'Stick2')]

    extra_optimization_maps = [
        lambda results: {'FS': 1 - results['w_ball.w']},
        lambda results: {'FS.std': results['w_ball.w.std']}
    ]
    extra_sampling_maps = [
        lambda samples: {
            'FS': np.mean(samples['w_stick0.w'] + samples['w_stick1.w'] + samples['w_stick2.w'], axis=1),
            'FS.std': np.std(samples['w_stick0.w'] + samples['w_stick1.w'] + samples['w_stick2.w'], axis=1)}
    ]

    extra_prior = 'return w_stick2.w < w_stick1.w && w_stick1.w < w_stick0.w;'



import os
from copy import copy
from textwrap import dedent

import yaml
from PyQt5.QtCore import pyqtSlot, QObject, pyqtSignal
from PyQt5.QtWidgets import QFileDialog, QMessageBox, QDialog, QDialogButtonBox

import mdt
import mot
from mdt.lib.components import get_meta_info, list_composite_models, list_cascade_models
from mdt.gui.model_fit.design.ui_fit_model_tab import Ui_FitModelTabContent
from mdt.gui.model_fit.design.ui_optimization_extra_data_add_protocol_map_dialog import Ui_AddProtocolMapDialog
from mdt.gui.model_fit.design.ui_optimization_extra_data_dialog import Ui_OptimizationExtraDataDialog
from mdt.gui.model_fit.design.ui_optimization_options_dialog import Ui_OptimizationOptionsDialog
from mdt.gui.utils import function_message_decorator, image_files_filters, protocol_files_filters, MainTab, \
    get_script_file_header_text
from mdt.utils import split_image_path
from mot.lib.cl_environments import CLEnvironmentFactory
from mot.optimize import get_minimizer_options

__author__ = 'Robbert Harms'
__date__ = "2016-06-27"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class FitModelTab(MainTab, Ui_FitModelTabContent, QObject):

    def __init__(self, shared_state, computations_thread):
        super().__init__()
        self._shared_state = shared_state
        self._computations_thread = computations_thread
        self._run_model_worker = RunModelWorker()
        self._tab_content = None
        self._optim_options = OptimOptions()
        self._input_data_info = InputDataInfo()

    def setupUi(self, tab_content):
        super().setupUi(tab_content)
        self._tab_content = tab_content

        self.selectDWI.clicked.connect(lambda: self._select_dwi())
        self.selectMask.clicked.connect(lambda: self._select_mask())
        self.selectProtocol.clicked.connect(lambda: self._select_protocol())
        self.selectOutputFolder.clicked.connect(lambda: self._select_output())

        self.selectedDWI.textChanged.connect(self._check_enable_action_buttons)
        self.selectedMask.textChanged.connect(self._check_enable_action_buttons)
        self.selectedProtocol.textChanged.connect(self._check_enable_action_buttons)

        self.runButton.clicked.connect(lambda: self.run_model())
        self.optimizationOptionsButton.clicked.connect(self._run_optimization_options_dialog)
        self.additionalDataButton.clicked.connect(self._extra_data_dialog)

        self.cascadedFitButtonGroup.buttonClicked.connect(self._update_cascade_selection_possible)

        self.modelSelection.addItems(list(sorted(list_composite_models())))
        self.modelSelection.currentIndexChanged.connect(self._update_cascade_selection)
        initial_model = 'BallStick_r1'

        self.modelSelection.setCurrentText(initial_model)
        self._update_cascade_selection()
        self.cascadeSelection.setCurrentText('Cascade')

        if self._input_data_info.dwi:
            self.selectedDWI.setText(self._input_data_info.dwi)
        if self._input_data_info.mask:
            self.selectedMask.setText(self._input_data_info.mask)
        if self._input_data_info.protocol:
            self.selectedProtocol.setText(self._input_data_info.protocol)

        self.update_output_folder_text()
        self._check_enable_action_buttons()

    def _update_cascade_selection_possible(self):
        self.cascadeSelection.setDisabled(self.noCascades.isChecked())

    def _update_cascade_selection(self):
        composite_model_name = self.modelSelection.currentText()
        cascade_models = list_cascade_models(composite_model_name)

        previous_cascade_selection = self.cascadeSelection.currentText()

        cascade_type_names = [get_meta_info('cascade_models', model)['cascade_type_name'] for model in cascade_models]

        self.cascadeSelection.clear()
        self.cascadeSelection.addItems(cascade_type_names)

        if previous_cascade_selection in cascade_type_names:
            self.cascadeSelection.setCurrentText(previous_cascade_selection)
        elif 'Cascade' in cascade_type_names:
            self.cascadeSelection.setCurrentText('Cascade')

        if not cascade_type_names:
            self.noCascades.setChecked(True)
            self._update_cascade_selection_possible()
            self.useCascades.setDisabled(True)
        else:
            self.useCascades.setEnabled(True)
            self.useCascades.setChecked(True)
            self._update_cascade_selection_possible()

    def _select_dwi(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedDWI.text() != '':
            initial_dir = self.selectedDWI.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the 4d (diffusion weighted) image', directory=initial_dir,
            filter=';;'.join(image_files_filters))

        if os.path.isfile(open_file):
            self.selectedDWI.setText(open_file)
            self._input_data_info.dwi = open_file
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.update_output_folder_text()

    def _select_mask(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedMask.text() != '':
            initial_dir = self.selectedMask.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the brain mask', directory=initial_dir,
            filter=';;'.join(image_files_filters))

        if os.path.isfile(open_file):
            self.selectedMask.setText(open_file)
            self._input_data_info.mask = open_file
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.update_output_folder_text()

    def _select_protocol(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedProtocol.text() != '':
            initial_dir = self.selectedProtocol.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the protocol', directory=initial_dir,
            filter=';;'.join(protocol_files_filters))

        if os.path.isfile(open_file):
            self.selectedProtocol.setText(open_file)
            self._input_data_info.protocol = open_file
            self._shared_state.base_dir = os.path.dirname(open_file)

    def _select_output(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedOutputFolder.text() != '':
            initial_dir = self.selectedOutputFolder.text()

        output_file_name = QFileDialog().getExistingDirectory(
            caption='Select the output folder', directory=initial_dir)

        if output_file_name:
            self.selectedOutputFolder.setText(output_file_name)

    def _check_enable_action_buttons(self):
        self.runButton.setEnabled(
            os.path.isfile(self.selectedDWI.text()) and
            os.path.isfile(self.selectedMask.text()) and
            os.path.isfile(self.selectedProtocol.text()) and
            self.selectedOutputFolder.text() != '')

    def update_output_folder_text(self):
        if os.path.isfile(self.selectedDWI.text()) and os.path.isfile(self.selectedMask.text()):
            folder_base = os.path.join(os.path.dirname(self.selectedDWI.text()), 'output',
                                       split_image_path(self.selectedMask.text())[1])
            self.selectedOutputFolder.setText(folder_base)

    def _run_optimization_options_dialog(self):
        dialog = OptimizationOptionsDialog(self._shared_state, self._tab_content, self._optim_options)
        return_value = dialog.exec_()

        if return_value:
            dialog.write_config()

    def _extra_data_dialog(self):
        dialog = ExtraDataDialog(self._shared_state, self._tab_content, self._input_data_info)
        return_value = dialog.exec_()

        if return_value:
            dialog.write_config()

    def _get_current_model_name(self):
        model_name = self.modelSelection.currentText()
        if self.useCascades.isChecked():
            model_name += ' ({})'.format(self.cascadeSelection.currentText())
        return model_name

    @pyqtSlot()
    def run_model(self):
        model = mdt.get_model(self._get_current_model_name())()

        if not model.is_input_data_sufficient(self._input_data_info.build_input_data()):
            msg = ProtocolWarningBox(model.get_input_data_problems(self._input_data_info.build_input_data()))
            msg.exec_()
            return

        self._run_model_worker.set_args(
            model,
            self._input_data_info.build_input_data(),
            self.selectedOutputFolder.text(),
            recalculate=True,
            double_precision=self._optim_options.double_precision,
            only_recalculate_last=not self._optim_options.recalculate_all,
            method=self._optim_options.method)

        self._computations_thread.start()
        self._run_model_worker.moveToThread(self._computations_thread)

        self._run_model_worker.starting.connect(self._computations_thread.starting)
        self._run_model_worker.finished.connect(self._computations_thread.finished)

        self._run_model_worker.starting.connect(lambda: self.runButton.setEnabled(False))
        self._run_model_worker.finished.connect(lambda: self.runButton.setEnabled(True))

        self._run_model_worker.finished.connect(
            lambda: self._shared_state.set_output_folder(self._get_full_model_output_path()))

        image_path = split_image_path(self._input_data_info.dwi)
        script_basename = os.path.join(image_path[0], 'scripts',
                                       'fit_model_{}_{}'.format(self._get_current_model_name().replace('|', '.'),
                                                                image_path[1]))
        if not os.path.isdir(os.path.join(image_path[0], 'scripts')):
            os.makedirs(os.path.join(image_path[0], 'scripts'))

        script_info = dict(optim_options=self._optim_options,
                           input_data_info=self._input_data_info,
                           model=self._get_current_model_name(),
                           output_folder=self.selectedOutputFolder.text(),
                           recalculate=True,
                           double_precision=self._optim_options.double_precision,
                           only_recalculate_last=not self._optim_options.recalculate_all)

        self._run_model_worker.finished.connect(
            lambda: self._write_python_script_file(script_basename + '.py', **script_info))

        self._run_model_worker.finished.connect(
            lambda: self._write_bash_script_file(script_basename + '.sh', **script_info))

        self._run_model_worker.starting.emit()

    def _get_full_model_output_path(self):
        parts = [self.selectedOutputFolder.text()]
        parts.append(self._get_current_model_name().split(' ')[0])
        return os.path.join(*parts)

    def _write_python_script_file(self, output_file, **kwargs):
        input_data_info = kwargs['input_data_info']
        optim_options = kwargs['optim_options']

        all_cl_devices = CLEnvironmentFactory.smart_device_selection()
        user_selected_devices = mot.configuration.get_cl_environments()

        format_kwargs = dict(
            header=get_script_file_header_text({'Purpose': 'Fitting a model'}),
            dwi=input_data_info.dwi,
            protocol=input_data_info.protocol,
            mask=input_data_info.mask,
            noise_std=input_data_info.noise_std,
            gradient_deviations=input_data_info.gradient_deviations,
            extra_protocol=input_data_info.extra_protocol,
            model=kwargs['model'],
            output_folder=kwargs['output_folder'],
            recalculate=kwargs['recalculate'],
            only_recalculate_last=kwargs['only_recalculate_last'],
            double_precision=kwargs['double_precision'],
            cl_device_ind=[ind for ind, device in enumerate(all_cl_devices) if device in user_selected_devices])

        with open(output_file, 'w') as f:
            f.write('#!/usr/bin/env python\n')

            if optim_options.use_model_default_optimizer:
                f.write(dedent('''
                    {header}

                    import mdt

                    input_data = mdt.load_input_data(
                        {dwi!r},
                        {protocol!r},
                        {mask!r},
                        noise_std={noise_std!r},
                        gradient_deviations={gradient_deviations!r},
                        extra_protocol={extra_protocol!r})

                    mdt.fit_model(
                        {model!r},
                        input_data,
                        {output_folder!r},
                        recalculate={recalculate!r},
                        only_recalculate_last={only_recalculate_last!r},
                        double_precision={double_precision!r},
                        cl_device_ind={cl_device_ind!r})

                ''').format(**format_kwargs))
            else:
                format_kwargs.update({'method': optim_options.method, 'patience': optim_options.patience})
                f.write(dedent('''
                    {header}

                    import mdt
                    from mdt.configuration import SetGeneralOptimizer

                    input_data = mdt.load_input_data(
                        {dwi!r},
                        {protocol!r},
                        {mask!r},
                        noise_std={noise_std!r},
                        gradient_deviations={gradient_deviations!r},
                        extra_protocol={extra_protocol!r})

                    with mdt.config_context(SetGeneralOptimizer({method!r}, settings={{'patience': {patience!r}}})):
                        mdt.fit_model(
                            {model!r},
                            input_data,
                            {output_folder!r},
                            recalculate={recalculate!r},
                            only_recalculate_last={only_recalculate_last!r},
                            double_precision={double_precision!r},
                            cl_device_ind={cl_device_ind!r})

                ''').format(**format_kwargs))

    def _write_bash_script_file(self, output_file, *args, **kwargs):
        input_data_info = kwargs['input_data_info']
        optim_options = kwargs['optim_options']

        all_cl_devices = CLEnvironmentFactory.smart_device_selection()
        user_selected_devices = mot.configuration.get_cl_environments()

        with open(output_file, 'w') as f:
            f.write('#!/usr/bin/env bash\n')
            f.write(dedent('''
                {header}

                mdt-model-fit \\
                    "{model}" \\
                    "{dwi}" \\
                    "{protocol}" \\
                    "{mask}" ''').format(header=get_script_file_header_text({'Purpose': 'Fitting a model'}),
                            model=kwargs['model'],
                            dwi=input_data_info.dwi,
                            protocol=input_data_info.protocol,
                            mask=input_data_info.mask))

            def write_new_line(line):
                f.write('\\\n' + ' ' * 4 + line + ' ')

            write_new_line('-o "{}"'.format(kwargs['output_folder']))

            if input_data_info.gradient_deviations:
                write_new_line('--gradient-deviations "{}"'.format(input_data_info.gradient_deviations))

            if input_data_info.noise_std:
                write_new_line('--noise-std {}'.format(input_data_info.noise_std))

            write_new_line('--cl-device-ind {}'.format(
                ' '.join(str(ind) for ind, device in enumerate(all_cl_devices) if device in user_selected_devices)))

            write_new_line('--recalculate' if kwargs['recalculate'] else '--no-recalculate')
            write_new_line('--only-recalculate-last' if kwargs['only_recalculate_last'] else '--recalculate-all')
            write_new_line('--double' if kwargs['double_precision'] else '--float')

            if input_data_info.extra_protocol:
                write_new_line('--extra-protocol {}'.format(
                    ' '.join('{}="{}"'.format(key, value) for key, value in input_data_info.extra_protocol.items())
                ))

            if not optim_options.use_model_default_optimizer:
                config = '''
                    optimization:
                        general:
                            name: {}
                            settings:
                                patience: {}
                '''.format(optim_options.method, optim_options.patience)
                config_context = yaml.safe_dump(yaml.safe_load(config), default_flow_style=True).rstrip()
                write_new_line('--config-context "{}"'.format(config_context))


class ProtocolWarningBox(QMessageBox):

    def __init__(self, problems, *args):
        super().__init__(*args)
        self.setIcon(QMessageBox.Warning)
        self.setWindowTitle("Insufficient protocol")
        self.setText("The provided protocol is insufficient for this model.")
        self.setInformativeText("The reported problems are: \n{}".format('\n'.join(' - ' + str(p) for p in problems)))
        self._in_resize = False

    def resizeEvent(self, event):
        if not self._in_resize:
            self._in_resize = True
            self.setFixedWidth(self.children()[-1].size().width() + 200)
            self._in_resize = False


class OptimizationOptionsDialog(Ui_OptimizationOptionsDialog, QDialog):

    def __init__(self, shared_state, parent, config):
        super().__init__(parent)
        self._shared_state = shared_state
        self._config = config
        self.setupUi(self)

        self.patience.textChanged.connect(self._check_enable_ok_button)

        self.optimizationRoutine.addItems(sorted(OptimOptions.optim_routines.keys()))
        self.optimizationRoutine.currentIndexChanged.connect(self._update_default_patience)
        self.defaultOptimizerGroup.buttonClicked.connect(self._update_optimization_routine_selection)

        self._load_config()

    def write_config(self):
        """Write to the config the user selected options"""
        self._config.double_precision = self.doublePrecision.isChecked()
        self._config.recalculate_all = self.recalculateAll_True.isChecked()
        self._config.use_model_default_optimizer = self.defaultOptimizer_True.isChecked()
        self._config.method = OptimOptions.optim_routines[self.optimizationRoutine.currentText()]
        self._config.patience = int(self.patience.text())

    def _load_config(self):
        """Load the settings from the config into the GUI"""
        self.doublePrecision.setChecked(self._config.double_precision)
        self.recalculateAll_True.setChecked(self._config.recalculate_all)
        self.defaultOptimizer_False.setChecked(not self._config.use_model_default_optimizer)
        self._update_optimization_routine_selection()

        self.optimizationRoutine.setCurrentText({v: k for k, v in
                                                 OptimOptions.optim_routines.items()}[self._config.method])
        self.patience.setText(str(self._config.patience))

    def _check_enable_ok_button(self):
        enabled = True
        if self.defaultOptimizer_False.isChecked():
            try:
                int(self.patience.text())
            except ValueError:
                enabled = False

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    def _update_optimization_routine_selection(self):
        self.optimizationRoutine.setDisabled(self.defaultOptimizer_True.isChecked())
        self.patience.setDisabled(self.defaultOptimizer_True.isChecked())

    def _update_default_patience(self):
        method = OptimOptions.optim_routines[self.optimizationRoutine.currentText()]
        self.patience.setText(str(get_minimizer_options(method)['patience']))


class OptimOptions(object):

    optim_routines = {'Powell\'s method': 'Powell',
                      'Nelder-Mead Simplex': 'Nelder-Mead',
                      'Levenberg Marquardt': 'Levenberg-Marquardt'}

    def __init__(self):
        """Storage class for communication between the options dialog and the main frame"""
        self.use_model_default_optimizer = True
        self.double_precision = False

        self.method = mdt.configuration.get_general_optimizer_name()
        self.patience = mdt.configuration.get_general_optimizer_options()['patience']

        if self.patience is None:
            self.patience = get_minimizer_options(self.method)['patience']

        self.recalculate_all = False


class InputDataInfo(object):

    def __init__(self):
        self.dwi = None
        self.mask = None
        self.protocol = None
        self.extra_protocol = {}
        self.gradient_deviations = None
        self.noise_std = None

    def build_input_data(self):
        return mdt.load_input_data(self.dwi, self.protocol, self.mask,
                                   noise_std=self.noise_std, gradient_deviations=self.gradient_deviations,
                                   extra_protocol=self.extra_protocol)


class ExtraDataDialog(Ui_OptimizationExtraDataDialog, QDialog):

    def __init__(self, shared_state, parent, input_data_info):
        super().__init__(parent)
        self._shared_state = shared_state
        self._input_data_info = input_data_info
        self.setupUi(self)

        self.noiseStdFileSelect.clicked.connect(lambda: self._select_std_file())
        self.noiseStd.textChanged.connect(self._check_enable_ok_button)

        self.gradDevFileSelect.clicked.connect(lambda: self._select_grad_dev_file())
        self.gradientDeviations.textChanged.connect(self._check_enable_ok_button)

        self.addProtocolMap.clicked.connect(lambda: self._add_map_dialog())
        self.removeProtocolMap.clicked.connect(lambda: self._remove_selected_extra_protocol())

        self._extra_protocol = {}

        self._load_config()

    def _add_map_dialog(self):
        dialog = AddProtocolMapDialog(self._shared_state, self, self._extra_protocol)
        return_value = dialog.exec_()

        if return_value:
            dialog.write_config()
            self._update_map_view()

    def _remove_selected_extra_protocol(self):
        to_remove = [el.text() for el in self.protocolMaps.selectedItems()]
        extra_protocol = {}
        for key, value in self._extra_protocol.items():
            if '{}: {}'.format(key, value) not in to_remove:
                extra_protocol.update({key: value})
        self._extra_protocol = extra_protocol
        self._update_map_view()

    def write_config(self):
        """Write to the config the user selected options"""
        noise_std_value = self.noiseStd.text()
        if noise_std_value == '':
            self._input_data_info.noise_std = None
        else:
            self._input_data_info.noise_std = noise_std_value
            try:
                self._input_data_info.noise_std = float(noise_std_value)
            except ValueError:
                pass

        if self.gradientDeviations.text() == '':
            self._input_data_info.gradient_deviations = None
        else:
            self._input_data_info.gradient_deviations = self.gradientDeviations.text()

        self._input_data_info.extra_protocol = copy(self._extra_protocol)

    def _load_config(self):
        """Load the settings from the config into the GUI"""
        if self._input_data_info.noise_std is not None:
            self.noiseStd.setText(str(self._input_data_info.noise_std))
        if self._input_data_info.gradient_deviations is not None:
            self.gradientDeviations.setText(str(self._input_data_info.gradient_deviations))
        self._extra_protocol = copy(self._input_data_info.extra_protocol)
        self._update_map_view()

    def _select_std_file(self):
        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select a noise std volume', directory=self._shared_state.base_dir,
            filter=';;'.join(image_files_filters))

        if open_file:
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.noiseStd.setText(open_file)

    def _select_grad_dev_file(self):
        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select a gradient deviations volume', directory=self._shared_state.base_dir,
            filter=';;'.join(image_files_filters))

        if open_file:
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.gradientDeviations.setText(open_file)

    def _check_enable_ok_button(self):
        noise_std_value = self.noiseStd.text()
        noise_std_value_is_float = False
        try:
            float(noise_std_value)
            noise_std_value_is_float = True
        except ValueError:
            pass

        enabled = noise_std_value == '' or noise_std_value_is_float or os.path.isfile(noise_std_value)

        if self.gradientDeviations.text() != '' and not os.path.isfile(self.gradientDeviations.text()):
            enabled = False

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    def _update_map_view(self):
        self.protocolMaps.clear()
        self.protocolMaps.addItems('{}: {}'.format(key, self._extra_protocol[key])
                                   for key in sorted(self._extra_protocol))


class AddProtocolMapDialog(Ui_AddProtocolMapDialog, QDialog):

    def __init__(self, shared_state, parent, extra_protocol):
        super().__init__(parent)
        self._shared_state = shared_state
        self._extra_protocol = extra_protocol
        self.setupUi(self)

        self.mapNameInput.textChanged.connect(self._check_enable_ok_button)
        self.valueInput.textChanged.connect(self._check_enable_ok_button)

        self.fileBrowse.clicked.connect(lambda: self._select_value_file())

        self._check_enable_ok_button()

    def write_config(self):
        """Write to the config the user selected options"""
        value = self.valueInput.text()
        try:
            value = float(value)
        except ValueError:
            pass

        self._extra_protocol.update({self.mapNameInput.text(): value})

    def _select_value_file(self):
        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select a noise std volume', directory=self._shared_state.base_dir,
            filter=';;'.join(image_files_filters))

        if open_file:
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.valueInput.setText(open_file)

    def _check_enable_ok_button(self):
        value = self.valueInput.text()
        value_is_float = False
        try:
            float(value)
            value_is_float = True
        except ValueError:
            pass

        enabled = value_is_float or os.path.isfile(value)

        if self.mapNameInput.text() == '':
            enabled = False

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)


class RunModelWorker(QObject):

    starting = pyqtSignal()
    finished = pyqtSignal()

    def __init__(self):
        super().__init__()
        self.starting.connect(self.run)
        self._args = []
        self._kwargs = {}

    def set_args(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    @function_message_decorator('Starting model fitting, please wait.',
                                'Finished model fitting. You can view the results using the "View results" tab.')
    @pyqtSlot()
    def run(self):
        mdt.fit_model(*self._args, **self._kwargs)
        self.finished.emit()




import glob
import os
import numpy as np
from mdt.utils import restore_volumes
from mdt.lib.nifti import write_nifti, load_nifti

__author__ = 'Robbert Harms'
__date__ = "2017-02-28"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def volume_map_npy_to_nifti(npy_fname, nifti_header, nifti_fname=None):
    """Convert a volume-map npy file to a nifti file.

    Args:
        npy_fname (str): the filename of the npy file to load
        nifti_header (nibabel header): the header file for the nifti
        nifti_fname (str): the filename of the nifti file. If not given it defaults to the same directory as the
            npy file.
    """
    data = np.load(npy_fname, mmap_mode='r')

    if nifti_fname is None:
        nifti_fname = os.path.join(os.path.dirname(npy_fname),
                                   os.path.splitext(os.path.basename(npy_fname))[0] + '.nii.gz')
    write_nifti(data, nifti_fname, nifti_header)


def load_all_npy_files(directory):
    """Load all the npy files in the given directory.

    Args:
        directory (str): the directory to load the npy files in

    Returns:
        dict: the loaded npy files with as keys the filename (without the .npy extension) and as value the
            memory mapped array
    """
    file_names = list(map(lambda p: os.path.splitext(os.path.basename(p))[0],
                          glob.glob(os.path.join(directory, '*.npy'))))

    results_dict = {}
    for file_name in file_names:
        results_dict[file_name] = np.load(os.path.join(directory, file_name + '.npy'), mmap_mode='r')
    return results_dict


def samples_npy_to_nifti(samples_npy_fname, used_mask, nifti_header, nifti_fname=None):
    """Convert a npy file containing sample results to a nifti file.

    Since the sample npy files are stored as a two dimensional matrix (with on the first axis the ROI index number
    and on the second the samples), we need to have the lookup table for the spatial information about the samples.

    Args:
        samples_npy_fname (str): the filename of the samples file to convert
        used_mask (ndarray or str): either an three dimensional matrix with the mask or a path to a nifti file.
        nifti_header (nibabel header): the header to use for writing the nifti file
        nifti_fname (str): the filename of the nifti file. If not given it defaults to the same directory as the
            samples file.
    """
    samples = np.load(samples_npy_fname, mmap_mode='r')

    if isinstance(used_mask, str):
        used_mask = load_nifti(used_mask).get_data()

    if np.count_nonzero(used_mask) != samples.shape[0]:
        raise ValueError(
            'The number of voxels in the mask ({}) does not correspond '
            'with the number of voxels in the samples file ({})'.format(np.count_nonzero(used_mask), samples.shape[0]))

    if nifti_fname is None:
        nifti_fname = os.path.join(os.path.dirname(samples_npy_fname),
                                   os.path.splitext(os.path.basename(samples_npy_fname))[0] + '.nii.gz')

    volume = restore_volumes(samples, used_mask)
    write_nifti(volume, nifti_fname, nifti_header)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'view_results_tab.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ViewResultsTabContent(object):
    def setupUi(self, ViewResultsTabContent):
        ViewResultsTabContent.setObjectName("ViewResultsTabContent")
        ViewResultsTabContent.resize(938, 427)
        self.verticalLayout = QtWidgets.QVBoxLayout(ViewResultsTabContent)
        self.verticalLayout.setContentsMargins(-1, 11, -1, -1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(ViewResultsTabContent)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.label_2 = QtWidgets.QLabel(ViewResultsTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.verticalLayout.addLayout(self.verticalLayout_2)
        self.line = QtWidgets.QFrame(ViewResultsTabContent)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setLineWidth(1)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout.setHorizontalSpacing(10)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.deselectAllButton = QtWidgets.QPushButton(ViewResultsTabContent)
        self.deselectAllButton.setObjectName("deselectAllButton")
        self.horizontalLayout_3.addWidget(self.deselectAllButton)
        self.invertSelectionButton = QtWidgets.QPushButton(ViewResultsTabContent)
        self.invertSelectionButton.setObjectName("invertSelectionButton")
        self.horizontalLayout_3.addWidget(self.invertSelectionButton)
        self.gridLayout.addLayout(self.horizontalLayout_3, 3, 1, 1, 1)
        self.label_3 = QtWidgets.QLabel(ViewResultsTabContent)
        self.label_3.setMinimumSize(QtCore.QSize(0, 0))
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 2, 0, 1, 1)
        self.selectMaps = QtWidgets.QListWidget(ViewResultsTabContent)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.selectMaps.sizePolicy().hasHeightForWidth())
        self.selectMaps.setSizePolicy(sizePolicy)
        self.selectMaps.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.selectMaps.setObjectName("selectMaps")
        self.gridLayout.addWidget(self.selectMaps, 2, 1, 1, 1)
        self.line_4 = QtWidgets.QFrame(ViewResultsTabContent)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout.addWidget(self.line_4, 1, 0, 1, 3)
        self.label_5 = QtWidgets.QLabel(ViewResultsTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 2, 2, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.selectFolderButton = QtWidgets.QPushButton(ViewResultsTabContent)
        self.selectFolderButton.setObjectName("selectFolderButton")
        self.horizontalLayout_2.addWidget(self.selectFolderButton)
        self.selectedFolderText = QtWidgets.QLineEdit(ViewResultsTabContent)
        self.selectedFolderText.setText("")
        self.selectedFolderText.setObjectName("selectedFolderText")
        self.horizontalLayout_2.addWidget(self.selectedFolderText)
        self.gridLayout.addLayout(self.horizontalLayout_2, 0, 1, 1, 1)
        self.label_6 = QtWidgets.QLabel(ViewResultsTabContent)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 0, 1, 1)
        self.label_4 = QtWidgets.QLabel(ViewResultsTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 0, 2, 1, 1)
        self.line_3 = QtWidgets.QFrame(ViewResultsTabContent)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout.addWidget(self.line_3, 4, 0, 1, 3)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setContentsMargins(0, -1, 0, -1)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem)
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setContentsMargins(-1, -1, 6, -1)
        self.horizontalLayout_9.setSpacing(0)
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.label_7 = QtWidgets.QLabel(ViewResultsTabContent)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_9.addWidget(self.label_7)
        self.horizontalLayout_4.addLayout(self.horizontalLayout_9)
        self.initialDimensionChooser = QtWidgets.QSpinBox(ViewResultsTabContent)
        self.initialDimensionChooser.setMaximum(2)
        self.initialDimensionChooser.setProperty("value", 2)
        self.initialDimensionChooser.setObjectName("initialDimensionChooser")
        self.horizontalLayout_4.addWidget(self.initialDimensionChooser)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem1)
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_10.setContentsMargins(-1, -1, 6, -1)
        self.horizontalLayout_10.setSpacing(0)
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.label_9 = QtWidgets.QLabel(ViewResultsTabContent)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_10.addWidget(self.label_9)
        self.horizontalLayout_4.addLayout(self.horizontalLayout_10)
        self.initialSliceChooser = QtWidgets.QSpinBox(ViewResultsTabContent)
        self.initialSliceChooser.setObjectName("initialSliceChooser")
        self.horizontalLayout_4.addWidget(self.initialSliceChooser)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setContentsMargins(3, -1, -1, -1)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_8 = QtWidgets.QLabel(ViewResultsTabContent)
        self.label_8.setObjectName("label_8")
        self.horizontalLayout_5.addWidget(self.label_8)
        self.maximumIndexLabel = QtWidgets.QLabel(ViewResultsTabContent)
        self.maximumIndexLabel.setObjectName("maximumIndexLabel")
        self.horizontalLayout_5.addWidget(self.maximumIndexLabel)
        self.horizontalLayout_4.addLayout(self.horizontalLayout_5)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem2)
        self.gridLayout.addLayout(self.horizontalLayout_4, 5, 1, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.line_2 = QtWidgets.QFrame(ViewResultsTabContent)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout.addWidget(self.line_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, 6, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.viewButton = QtWidgets.QPushButton(ViewResultsTabContent)
        self.viewButton.setObjectName("viewButton")
        self.horizontalLayout.addWidget(self.viewButton)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem3)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.line_2.raise_()
        self.line.raise_()

        self.retranslateUi(ViewResultsTabContent)
        QtCore.QMetaObject.connectSlotsByName(ViewResultsTabContent)
        ViewResultsTabContent.setTabOrder(self.selectFolderButton, self.selectedFolderText)
        ViewResultsTabContent.setTabOrder(self.selectedFolderText, self.selectMaps)
        ViewResultsTabContent.setTabOrder(self.selectMaps, self.deselectAllButton)
        ViewResultsTabContent.setTabOrder(self.deselectAllButton, self.invertSelectionButton)
        ViewResultsTabContent.setTabOrder(self.invertSelectionButton, self.initialDimensionChooser)
        ViewResultsTabContent.setTabOrder(self.initialDimensionChooser, self.initialSliceChooser)
        ViewResultsTabContent.setTabOrder(self.initialSliceChooser, self.viewButton)

    def retranslateUi(self, ViewResultsTabContent):
        _translate = QtCore.QCoreApplication.translate
        ViewResultsTabContent.setWindowTitle(_translate("ViewResultsTabContent", "Form"))
        self.label.setText(_translate("ViewResultsTabContent", "View results"))
        self.label_2.setText(_translate("ViewResultsTabContent", "View a selection of maps in the given folder."))
        self.deselectAllButton.setText(_translate("ViewResultsTabContent", "Deselect all"))
        self.invertSelectionButton.setText(_translate("ViewResultsTabContent", "Invert selection"))
        self.label_3.setText(_translate("ViewResultsTabContent", "Select maps:"))
        self.label_5.setText(_translate("ViewResultsTabContent", "(Select the maps you would like to display)"))
        self.selectFolderButton.setText(_translate("ViewResultsTabContent", "Browse"))
        self.label_6.setText(_translate("ViewResultsTabContent", "Select input folder:"))
        self.label_4.setText(_translate("ViewResultsTabContent", "(Choose a directory with .nii(.gz) files)"))
        self.label_7.setText(_translate("ViewResultsTabContent", "Initial dimension:"))
        self.label_9.setText(_translate("ViewResultsTabContent", "Initial slice:"))
        self.label_8.setText(_translate("ViewResultsTabContent", "/ "))
        self.maximumIndexLabel.setText(_translate("ViewResultsTabContent", "x"))
        self.viewButton.setText(_translate("ViewResultsTabContent", "View"))




from mdt.models.base import NamedProtocolProblem
from mdt.models.base import DMRIOptimizable

__author__ = 'Robbert Harms'
__date__ = "2015-04-24"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DMRICascadeModelInterface(DMRIOptimizable):

    def __init__(self, *args, **kwargs):
        """The interface to cascade models.

        A cascade model is a model consisting of multi-compartment models or other cascade models. The idea is that
        it contains a number of models that are to be optimized one after another with the output results of
        the previous fit used to initialize the next model.
        """
        super().__init__(*args, **kwargs)

    @property
    def name(self):
        """Get the name of this cascade model.

        Returns:
            str: The name of this cascade model
        """
        return ''

    def has_next(self):
        """Check if this cascade model has a next model.

        Returns:
            boolean: True if there is a next model, false otherwise.
        """
        raise NotImplementedError()

    def get_next(self, output_previous_models):
        """Get the next model in the cascade. This is the only function called by the model optimizer.

        This cascade class is supposed to remember which model is next.

        Args:
            output_previous_models (list): The output of all the previous models. Each element are the results
                of the optimization run of one of the models in the cascade.

        Returns:
            DMRIOptimizable: The model used for the next fit
        """
        raise NotImplementedError()

    def reset(self):
        """Reset the iteration over the cascade.

        The implementing class should now reset the iteration such that get_next gets the first model again.
        """
        raise NotImplementedError()

    def get_model(self, name):
        """Get one of the models in the cascade by name.

        Args:
            name (str): the name of the model we want to return

        Returns:
            the model we want to have or None if no model found
        """
        raise NotImplementedError()

    def get_model_names(self):
        """Get the names of the models in this cascade in order of execution.

        Returns:
            list of str: the names of the models in this list
        """
        raise NotImplementedError()

    def set_input_data(self, input_data):
        """Set the input data in every model in the cascade."""
        raise NotImplementedError()

    def update_active_post_processing(self, processing_type, settings):
        """Update the active post-processing semaphores for every model.
        """
        raise NotImplementedError()


class SimpleCascadeModel(DMRICascadeModelInterface):

    def __init__(self, name, model_list):
        """Create a new cascade model from a given list of models.

        This class adds some standard bookkeeping to make implementing cascade models easier.

        Args:
            name (str): the name of this cascade model
            model_list (list of models): the list of models this cascade consists of
        """
        super().__init__()
        self._name = name
        self._model_list = model_list
        self._iteration_position = 0

        model_names = [model.name for model in model_list]
        if len(model_names) > len(set(model_names)):
            raise ValueError('Non-unique model names detected in the cascade. '
                             'Please ensure all models are uniquely named.')

    @property
    def name(self):
        return self._name

    def has_next(self):
        return self._iteration_position != len(self._model_list)

    def get_next(self, output_previous_models):
        next_model = self._model_list[self._iteration_position]

        output_previous = {}
        if self._iteration_position > 0:
            output_previous = output_previous_models[self._iteration_position - 1]

        self._prepare_model(self._iteration_position, next_model, output_previous, output_previous_models)
        self._iteration_position += 1
        return self._set_model_options(next_model)

    def reset(self):
        self._iteration_position = 0

    def is_input_data_sufficient(self, input_data=None):
        return all(model.is_input_data_sufficient(input_data) for model in self._model_list)

    def get_input_data_problems(self, input_data=None):
        problems = []
        for model in self._model_list:
            problems.extend(map(lambda p: NamedProtocolProblem(p, model.name), model.get_input_data_problems(input_data)))
        return problems

    def get_required_protocol_names(self):
        protocol_names = []
        for model in self._model_list:
            protocol_names.extend(model.get_required_protocol_names())
        return list(set(protocol_names))

    def get_model(self, name):
        for model in self._model_list:
            if model.name == name:
                return self._set_model_options(model)
        return None

    def get_model_names(self):
        return [model.name for model in self._model_list]

    def set_input_data(self, input_data):
        for model in self._model_list:
            model.set_input_data(input_data)

    def update_active_post_processing(self, processing_type, settings):
        for model in self._model_list:
            model.update_active_post_processing(processing_type, settings)

    def _set_model_options(self, model):
        """The final hook before we return a model from this class.

        This can set all kind of additional extra's to the model before we return it using any of the functions
        in this class

        Args:
            model: the model to which we want to set the final functions

        Returns:
            model: the same model with all extra's set.
        """
        return model

    def _prepare_model(self, iteration_position, model, output_previous, output_all_previous):
        """Prepare the next model with the output of the previous model.

        By default this model initializes all parameter maps to the output of the previous model.

        Args:
            iteration_position (int): the index (in the list of cascades) of the model we are initializing.
                First model has position 0, then 1 etc.
            model: The model to prepare
            output_previous (dict): the output of the (direct) previous model.
            output_all_previous (list): The output of all the previous models. Every element (indexed by position in the
                cascade) contains the full set of results from the optimization of that specific model.

        Returns:
            None, preparing should happen in-place.
        """
        if not isinstance(model, DMRICascadeModelInterface) and output_previous:
            for key, value in output_previous.items():
                if model.has_parameter(key):
                    model.init(key, value)





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'generate_protocol_update_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_UpdateColumnDialog(object):
    def setupUi(self, UpdateColumnDialog):
        UpdateColumnDialog.setObjectName("UpdateColumnDialog")
        UpdateColumnDialog.resize(831, 304)
        self.verticalLayout = QtWidgets.QVBoxLayout(UpdateColumnDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(UpdateColumnDialog)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(UpdateColumnDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_3.addWidget(self.label_4)
        self.verticalLayout.addLayout(self.verticalLayout_3)
        self.line = QtWidgets.QFrame(UpdateColumnDialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.fileInput = QtWidgets.QPushButton(UpdateColumnDialog)
        self.fileInput.setEnabled(True)
        self.fileInput.setObjectName("fileInput")
        self.horizontalLayout.addWidget(self.fileInput)
        self.selectedFile = QtWidgets.QLineEdit(UpdateColumnDialog)
        self.selectedFile.setEnabled(True)
        self.selectedFile.setObjectName("selectedFile")
        self.horizontalLayout.addWidget(self.selectedFile)
        self.gridLayout.addLayout(self.horizontalLayout, 4, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(UpdateColumnDialog)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)
        self.inputMethodSelector = QtWidgets.QComboBox(UpdateColumnDialog)
        self.inputMethodSelector.setObjectName("inputMethodSelector")
        self.inputMethodSelector.addItem("")
        self.inputMethodSelector.addItem("")
        self.gridLayout.addWidget(self.inputMethodSelector, 1, 1, 1, 1)
        self.label = QtWidgets.QLabel(UpdateColumnDialog)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.valueScale = QtWidgets.QLineEdit(UpdateColumnDialog)
        self.valueScale.setObjectName("valueScale")
        self.gridLayout.addWidget(self.valueScale, 6, 1, 1, 1)
        self.label_11 = QtWidgets.QLabel(UpdateColumnDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_11.setFont(font)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 3, 2, 1, 1)
        self.label_10 = QtWidgets.QLabel(UpdateColumnDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_10.setFont(font)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 0, 2, 1, 1)
        self.label_7 = QtWidgets.QLabel(UpdateColumnDialog)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 6, 0, 1, 1)
        self.label_5 = QtWidgets.QLabel(UpdateColumnDialog)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 3, 0, 1, 1)
        self.columnNameInput = QtWidgets.QLineEdit(UpdateColumnDialog)
        self.columnNameInput.setObjectName("columnNameInput")
        self.gridLayout.addWidget(self.columnNameInput, 0, 1, 1, 1)
        self.label_12 = QtWidgets.QLabel(UpdateColumnDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_12.setFont(font)
        self.label_12.setObjectName("label_12")
        self.gridLayout.addWidget(self.label_12, 4, 2, 1, 1)
        self.label_13 = QtWidgets.QLabel(UpdateColumnDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_13.setFont(font)
        self.label_13.setObjectName("label_13")
        self.gridLayout.addWidget(self.label_13, 6, 2, 1, 1)
        self.line_2 = QtWidgets.QFrame(UpdateColumnDialog)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout.addWidget(self.line_2, 2, 0, 1, 3)
        self.label_6 = QtWidgets.QLabel(UpdateColumnDialog)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 4, 0, 1, 1)
        self.singleValueInput = QtWidgets.QLineEdit(UpdateColumnDialog)
        self.singleValueInput.setObjectName("singleValueInput")
        self.gridLayout.addWidget(self.singleValueInput, 3, 1, 1, 1)
        self.line_4 = QtWidgets.QFrame(UpdateColumnDialog)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout.addWidget(self.line_4, 5, 0, 1, 3)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.line_3 = QtWidgets.QFrame(UpdateColumnDialog)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.buttonBox = QtWidgets.QDialogButtonBox(UpdateColumnDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(UpdateColumnDialog)
        self.buttonBox.accepted.connect(UpdateColumnDialog.accept)
        self.buttonBox.rejected.connect(UpdateColumnDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(UpdateColumnDialog)
        UpdateColumnDialog.setTabOrder(self.columnNameInput, self.inputMethodSelector)
        UpdateColumnDialog.setTabOrder(self.inputMethodSelector, self.singleValueInput)
        UpdateColumnDialog.setTabOrder(self.singleValueInput, self.fileInput)
        UpdateColumnDialog.setTabOrder(self.fileInput, self.selectedFile)
        UpdateColumnDialog.setTabOrder(self.selectedFile, self.valueScale)

    def retranslateUi(self, UpdateColumnDialog):
        _translate = QtCore.QCoreApplication.translate
        UpdateColumnDialog.setWindowTitle(_translate("UpdateColumnDialog", "Add / Update column"))
        self.label_3.setText(_translate("UpdateColumnDialog", "Add / Update column"))
        self.label_4.setText(_translate("UpdateColumnDialog", "Add a column to the current protocol or overwrite an existing column."))
        self.fileInput.setText(_translate("UpdateColumnDialog", "Browse"))
        self.label_2.setText(_translate("UpdateColumnDialog", "Method:"))
        self.inputMethodSelector.setItemText(0, _translate("UpdateColumnDialog", "From file"))
        self.inputMethodSelector.setItemText(1, _translate("UpdateColumnDialog", "Single value"))
        self.label.setText(_translate("UpdateColumnDialog", "Column name: "))
        self.label_11.setText(_translate("UpdateColumnDialog", "(A single value for every row)"))
        self.label_10.setText(_translate("UpdateColumnDialog", "(The column name, for example \"g\", \"b\" or \"TE\")"))
        self.label_7.setText(_translate("UpdateColumnDialog", "Scale:"))
        self.label_5.setText(_translate("UpdateColumnDialog", "Single value:"))
        self.label_12.setText(_translate("UpdateColumnDialog", "(File with a single value, a row, a column or a matrix)"))
        self.label_13.setText(_translate("UpdateColumnDialog", "(Optionally, scale the input with this amount)"))
        self.label_6.setText(_translate("UpdateColumnDialog", "File input:"))




"""Contains the runtime configuration of MDT.

This consists of two parts, functions to get the current runtime settings and configuration actions to update these
settings. To set a new configuration, create a new :py:class:`ConfigAction` and use this within a context environment
using :py:func:`config_context`. Example:

.. code-block:: python

    from mdt.configuration import YamlStringAction, config_context

    config = '''
        optimization:
            general:
                name: 'Powell'
                settings:
                    patience: 2
    '''
    with mdt.config_context(YamlStringAction(config)):
        mdt.fit_model(...)
"""
import os
import re
from copy import deepcopy

import collections
import yaml
from contextlib import contextmanager
from pkg_resources import resource_stream

import mot.configuration
from mot.lib.load_balance_strategies import EvenDistribution

from mdt.__version__ import __version__

__author__ = 'Robbert Harms'
__date__ = "2015-06-23"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"

""" The current configuration """
_config = {}


def get_config_option(option_name):
    """Get the current configuration option for the given option name.

    Args:
        option_name (list of str or str): the name of the option, or a path to the option.

    Returns:
        object: the raw configuration value defined for that option
    """
    if isinstance(option_name, str):
        return _config[option_name]
    else:
        config = _config
        for el in option_name[:-1]:
            config = config[el]
        return config[option_name[-1]]


def set_config_option(option_name, value):
    """Set the current configuration option for the given option name.

    This will overwrite the current configuration for that option with the given value. Be careful, this will change
    the global configuration value.

    Provided values should be objects and not YAML strings. For updating the configuration with YAML strings, please use
    the function :func:`load_from_yaml`.

    Args:
        option_name (list of str or str): the name of the option, or a path to the option.
        value (object): the object to set for that option

    Returns:
        object: the raw configuration value defined for that option
    """
    if isinstance(option_name, str):
        _config[option_name] = value
    else:
        config = _config
        for el in option_name[:-1]:
            config = config[el]
        config[option_name[-1]] = value


def get_config_dir():
    """Get the location of the components.

    Return:
        str: the path to the components
    """
    return os.path.join(os.path.expanduser("~"), '.mdt', __version__)


def _config_insert(keys, value):
    """Insert the given value in the given key.

    This will create all layers of the dictionary if needed.

    Args:
        keys (list of str): the position of the input value
        value (object): the value to put at the position of the key.
    """
    config = _config
    for key in keys[:-1]:
        if key not in config:
            config[key] = {}
        config = config[key]

    config[keys[-1]] = value


def ensure_exists(keys):
    """Ensure the given layer of keys exists.

    Args:
        keys (list of str): the positions to ensure exist
    """
    config = _config
    for key in keys:
        if key not in config:
            config[key] = {}
        config = config[key]


def load_builtin():
    """Load the config file from the skeleton in mdt/data/mdt.conf"""
    with resource_stream('mdt', 'data/mdt.conf') as f:
        load_from_yaml(f.read())


def load_user_home():
    """Load the config file from the user home directory"""
    config_file = os.path.join(get_config_dir(), 'mdt.conf')
    if os.path.isfile(config_file):
        with open(config_file) as f:
            load_from_yaml(f.read())
    else:
        raise IOError('Config file could not be loaded.')


def load_user_gui():
    """Load the gui specific config file from the user home directory"""
    config_file = os.path.join(get_config_dir(), 'mdt.gui.conf')
    if os.path.isfile(config_file):
        with open(config_file) as f:
            load_from_yaml(f.read())
    else:
        raise IOError('Config file could not be loaded.')


def load_specific(file_name):
    """Can be called by the application to use the config from a specific file.

    This assumes that the given file contains YAML content, that is, we want to process it
    with the function load_from_yaml().

    Please note that the last configuration loaded overwrites the values of the previously loaded config files.

    Also, please note that this will change the global configuration, i.e. this is a persistent change. If you do not
    want a persistent state change, consider using :func:`~mdt.configuration.config_context` instead.

    Args:
        file_name (str): The name of the file to use.
    """
    with open(file_name) as f:
        load_from_yaml(f.read())


def load_from_yaml(yaml_str):
    """Can be called to use configuration options from a YAML string.

    This will update the current configuration with the new options.

    Please note that this will change the global configuration, i.e. this is a persistent change. If you do not
    want a persistent state change, consider using :func:`~mdt.configuration.config_context` instead.

    Args:
        yaml_str (str): The string containing the YAML config to parse.
    """
    config_dict = yaml.safe_load(yaml_str) or {}
    load_from_dict(config_dict)


def load_from_dict(config_dict):
    """Load configuration options from a given dictionary.

    Please note that this will change the global configuration, i.e. this is a persistent change. If you do not
    want a persistent state change, consider using :func:`~mdt.configuration.config_context` instead.

    Args:
        config_dict (dict): the dictionary from which to use the configurations
    """
    for key, value in config_dict.items():
        loader = get_section_loader(key)
        loader.load(value)


def update_gui_config(update_dict):
    """Update the GUI configuration file with the given settings.

    Args:
        update_dict (dict): the items to update in the GUI config file
    """
    update_write_config(os.path.join(get_config_dir(), 'mdt.gui.conf'), update_dict)


def update_write_config(config_file, update_dict):
    """Update a given configuration file with updated values.

    If the configuration file does not exist, a new one is created.

    Args:
        config_file (str): the location of the config file to update
        update_dict (dict): the items to update in the config file
    """
    if not os.path.exists(config_file):
        with open(config_file, 'a'):
            pass

    with open(config_file, 'r') as f:
        config_dict = yaml.safe_load(f.read()) or {}

    for key, value in update_dict.items():
        loader = get_section_loader(key)
        loader.update(config_dict, value)

    with open(config_file, 'w') as f:
        yaml.safe_dump(config_dict, f)


class ConfigSectionLoader(object):

    def load(self, value):
        """Load the given configuration value into the current configuration.

        Args:
            value: the value to use in the configuration
        """

    def update(self, config_dict, updates):
        """Update the given configuration dictionary with the values in the given updates dict.

        This enables automating updating a configuration file. Updates are written in place.

        Args:
            config_dict (dict): the current configuration dict
            updates (dict): the updated values to add to the given config dict.
        """


class OutputFormatLoader(ConfigSectionLoader):
    """Loader for the top level key output_format. """

    def load(self, value):
        for item in ['optimization', 'sampling']:
            options = value.get(item, {})

            if 'gzip' in options:
                _config_insert(['output_format', item, 'gzip'], bool(options['gzip']))


class LoggingLoader(ConfigSectionLoader):
    """Loader for the top level key logging. """

    def load(self, value):
        ensure_exists(['logging', 'info_dict'])
        if 'info_dict' in value:
            self._load_info_dict(value['info_dict'])

    def _load_info_dict(self, info_dict):
        for item in ['version', 'disable_existing_loggers', 'formatters', 'handlers', 'loggers', 'root']:
            if item in info_dict:
                _config_insert(['logging', 'info_dict', item], info_dict[item])


class OptimizationSettingsLoader(ConfigSectionLoader):
    """Loads the optimization section"""

    def load(self, value):
        ensure_exists(['optimization', 'general'])
        ensure_exists(['optimization', 'model_specific'])

        if 'general' in value:
            _config_insert(['optimization', 'general'], value['general'])

        if 'model_specific' in value:
            for key, sub_value in value['model_specific'].items():
                _config_insert(['optimization', 'model_specific', key], sub_value)


class SampleSettingsLoader(ConfigSectionLoader):
    """Loads the sample section"""

    def load(self, value):
        ensure_exists(['sampling', 'general'])

        settings = value.get('settings', {})
        settings['nmr_samples'] = settings.get('nmr_samples', 10000)
        settings['burnin'] = settings.get('burnin', 0)
        settings['thinning'] = settings.get('thinning', 1)
        _config_insert(['sampling', 'general', 'settings'], settings)


class ProcessingStrategySectionLoader(ConfigSectionLoader):
    """Loads the config section processing_strategies"""

    def load(self, value):
        if 'optimization' in value:
            _config_insert(['processing_strategies', 'optimization'], value['optimization'])
        if 'sampling' in value:
            _config_insert(['processing_strategies', 'sampling'], value['sampling'])


class TmpResultsDirSectionLoader(ConfigSectionLoader):
    """Load the section tmp_results_dir"""

    def load(self, value):
        _config_insert(['tmp_results_dir'], value)


class ActivePostProcessingLoader(ConfigSectionLoader):
    """Load the default settings for the post sample calculations."""

    def load(self, value):
        sampling = value.get('sampling', {})
        sampling['univariate_ess'] = sampling.get('univariate_ess', False)
        sampling['multivariate_ess'] = sampling.get('multivariate_ess', False)
        sampling['maximum_likelihood'] = sampling.get('maximum_likelihood', False)
        sampling['maximum_a_posteriori'] = sampling.get('maximum_a_posteriori', False)
        sampling['model_defined_maps'] = sampling.get('model_defined_maps', False)

        optimization = value.get('optimization', {})
        optimization['uncertainties'] = optimization.get('uncertainties', True)
        optimization['store_covariances'] = optimization.get('store_covariances', True)

        _config_insert(['active_post_processing', 'optimization'], optimization)
        _config_insert(['active_post_processing', 'sampling'], sampling)


class AutomaticCascadeModels(ConfigSectionLoader):
    """Load the automatic cascade model settings."""

    def load(self, value):
        _config_insert(['auto_generate_cascade_models', 'enabled'], value.get('enabled', True))
        _config_insert(['auto_generate_cascade_models', 'excluded'], value.get('excluded', []))


class RuntimeSettingsLoader(ConfigSectionLoader):

    def load(self, value):
        if 'cl_device_ind' in value:
            if value['cl_device_ind'] is not None:
                from mdt.utils import get_cl_devices
                all_devices = get_cl_devices()

                indices = value['cl_device_ind']
                if not isinstance(indices, collections.Iterable):
                    indices = [indices]

                devices = [all_devices[ind] for ind in indices if ind < len(all_devices)]

                if devices:
                    mot.configuration.set_cl_environments(devices)
                    mot.configuration.set_load_balancer(EvenDistribution())

    def update(self, config_dict, updates):
        if 'runtime_settings' not in config_dict:
            config_dict.update({'runtime_settings': {}})
        config_dict['runtime_settings'].update(updates)


def get_section_loader(section):
    """Get the section loader to use for the given top level section.

    Args:
        section (str): the section key we want to get the loader for

    Returns:
        ConfigSectionLoader: the config section loader for this top level section of the configuration.
    """
    if section == 'output_format':
        return OutputFormatLoader()

    if section == 'logging':
        return LoggingLoader()

    if section == 'optimization':
        return OptimizationSettingsLoader()

    if section == 'sampling':
        return SampleSettingsLoader()

    if section == 'processing_strategies':
        return ProcessingStrategySectionLoader()

    if section == 'tmp_results_dir':
        return TmpResultsDirSectionLoader()

    if section == 'runtime_settings':
        return RuntimeSettingsLoader()

    if section == 'auto_generate_cascade_models':
        return AutomaticCascadeModels()

    if section == 'active_post_processing':
        return ActivePostProcessingLoader()

    raise ValueError('Could not find a suitable configuration loader for the section {}.'.format(section))


def gzip_optimization_results():
    """Check if we should write the volume maps from the optimization gzipped or not.

    Returns:
        boolean: True if the results of optimization computations should be gzipped, False otherwise.
    """
    return _config['output_format']['optimization']['gzip']


def gzip_sampling_results():
    """Check if we should write the volume maps from the sample gzipped or not.

    Returns:
        boolean: True if the results of sample computations should be gzipped, False otherwise.
    """
    return _config['output_format']['sampling']['gzip']


def get_tmp_results_dir():
    """Get the default tmp results directory.

    This is the default directory for saving temporary computation results. Set to None to disable this and
    use the model directory.

    Returns:
        str or None: the tmp results dir to use during optimization and sample
    """
    return _config['tmp_results_dir']


def get_active_post_processing():
    """Get the overview of active post processing switches.

    Returns:
        dict: a dictionary holding two dictionaries, one called 'optimization' and one called 'sampling'.
            Both these dictionaries hold keys of elements to add to the respective post processing phase.
    """
    return deepcopy(_config['active_post_processing'])


def get_processing_strategy(processing_type, *args, **kwargs):
    """Get the correct processing strategy for the given model.

    Args:
        processing_type (str): 'optimization', 'sampling' or any other of the
            processing_strategies defined in the config
        model_names (list of str): the list of model names (the full recursive cascade of model names)
        **kwargs: passed to the constructor of the loaded processing strategy.

    Returns:
        ModelProcessingStrategy: the processing strategy to use for this model
    """
    from mdt.lib.processing_strategies import VoxelRange
    options = _config['processing_strategies'].get(processing_type, {}) or {}
    options.update(kwargs)
    return VoxelRange(*args, **options)


def get_logging_configuration_dict():
    """Get the configuration dictionary for the logging.dictConfig().

    MDT uses a few special logging configuration options to log to the files and GUI's. These options are defined
    using a configuration dictionary that this function returns.

    Returns:
        dict: the configuration dict for use with dictConfig of the Python logging modules
    """
    return _config['logging']['info_dict']


def get_general_optimizer():
    """Load the general optimizer from the configuration.

    Returns:
        Optimizer: the configured optimizer for use in MDT
    """
    return _config['optimization']['general']['name']


def get_optimizer_for_model(model_names):
    """Get the optimizer for this specific cascade of models.

    This configuration function supports having a different optimizer for optimizing, for example, NODDI in a
    cascade and NODDI without a cascade.

    Args:
        model_names (list of str): the list of model names (typically a cascade of models) for which we
            want to get the optimizer to use.

    Returns:
        Optimizer: the optimizer to use for optimizing the specific model
    """
    info_dict = get_model_config(model_names, _config['optimization']['model_specific'])

    if info_dict:
        return info_dict['name']
    else:
        return get_general_optimizer()


def get_general_optimizer_name():
    """Get the name of the currently configured general optimizer

    Returns:
        str: the name of the currently configured optimizer
    """
    return _config['optimization']['general']['name']


def get_general_optimizer_options():
    """Get the settings of the currently configured general optimizer

    Returns:
        dict: the settings of the currently configured optimizer
    """
    return _config['optimization']['general']['options']


def get_general_sampling_settings():
    """Get the general sample settings.

    Returns:
        Sampler: the configured sampler for use in MDT
    """
    return _config['sampling']['general']['settings']


def use_automatic_generated_cascades():
    """Check if we want to use the automatic cascade generation in MDT.

    Returns:
        boolean: True if we want to use the automatic cascades, False otherwise
    """
    return _config['auto_generate_cascade_models']['enabled']


def get_automatic_generated_cascades_excluded():
    """Get the information about the model names excluded from automatic cascade generation.

    Returns:
        list: the names of the composite models we want to exclude from automatic cascade generation
    """
    return _config['auto_generate_cascade_models']['excluded']


def get_model_config(model_names, config):
    """Get from the given dictionary the config for the given model.

    This tries to find the best match between the given config items (by key) and the given model list. For example
    if model_names is ['BallStick', 'S0'] and we have the following config dict:

    .. code-block:: python

        {'^S0$': 0,
         '^BallStick$': 1
         ('^BallStick$', '^S0$'): 2,
         ('^BallStickStick$', '^BallStick$', '^S0$'): 3,
         }

    then this function should return 2. because that one is the best match, even though the last option is also a
    viable match. That is, since a subset of the keys in the last option also matches the model names, it is
    considered a match as well. Still the best match is the third option (returning 2).

    Args:
        model_names (list of str): the names of the models we want to match. This should contain the entire
            recursive list of cascades leading to the composite model we want to get the config for.
        config (dict): the config items with as keys either a composite model regex for a name or a list of regex for
            a chain of model names.

    Returns:
        The config content of the best matching key.
    """
    if not config:
        return {}

    def get_key_length(key):
        if isinstance(key, tuple):
            return len(key)
        return 1

    def is_match(model_names, config_key):
        if isinstance(model_names, str):
            model_names = [model_names]

        if len(model_names) != get_key_length(config_key):
            return False

        if isinstance(config_key, tuple):
            return all([re.match(config_key[ind], model_names[ind]) for ind in range(len(config_key))])

        return re.match(config_key, model_names[0])

    key_length_lookup = ((get_key_length(key), key) for key in config.keys())
    ascending_keys = tuple(item[1] for item in sorted(key_length_lookup, key=lambda info: info[0]))

    # direct matching
    for key in ascending_keys:
        if is_match(model_names, key):
            return config[key]

    # partial matching string keys to last model name
    for key in ascending_keys:
        if not isinstance(key, tuple):
            if is_match(model_names[-1], key):
                return config[key]

    # partial matching tuple keys with a moving filter
    for key in ascending_keys:
        if isinstance(key, tuple):
            for start_ind in range(len(key)):
                sub_key = key[start_ind:]

                if is_match(model_names, sub_key):
                    return config[key]

    # no match found
    return {}


@contextmanager
def config_context(config_action):
    """Creates a temporary configuration context with the given config action.

    This will temporarily alter the given configuration keys to the given values. After the context is executed
    the configuration will revert to the original settings.

    Example usage:

    .. code-block:: python

        config = '''
            optimization:
                general:
                    name: 'Nelder-Mead'
                    options:
                        patience: 10
        '''
        with mdt.config_context(mdt.configuration.YamlStringAction(config)):
            mdt.fit_model(...)


    or, equivalently::

        config = '''
            ...
        '''
        with mdt.config_context(config):
            ...

    This loads the configuration from a YAML string and uses that configuration as the context.

    Args:
        config_action (ConfigAction or str): the configuration action to apply. If a string is given we will
            use it using the YamlStringAction config action.
    """
    if isinstance(config_action, str):
        config_action = YamlStringAction(config_action)

    config_action.apply()
    yield
    config_action.unapply()


class ConfigAction(object):

    def __init__(self):
        """Defines a configuration action for the use in a configuration context.

        This should define an apply and an unapply function that sets and unsets the given configuration options.

        The applying action needs to remember the state before applying the action.
        """

    def apply(self):
        """Apply the current action to the current runtime configuration."""

    def unapply(self):
        """Reset the current configuration to the previous state."""


class VoidConfigAction(ConfigAction):
    """Does nothing. Meant as a container to not have to check for None's everywhere."""

    def apply(self):
        pass

    def unapply(self):
        pass


class SimpleConfigAction(ConfigAction):

    def __init__(self):
        """Defines a default implementation of a configuration action.

        This simple config implements a default apply() method that saves the current state and a default
        unapply() that restores the previous state.

        It is easiest to implement _apply() for extra actions.
        """
        super().__init__()
        self._old_config = {}

    def apply(self):
        """Apply the current action to the current runtime configuration."""
        self._old_config = deepcopy(_config)
        self._apply()

    def unapply(self):
        """Reset the current configuration to the previous state."""
        global _config
        _config = self._old_config

    def _apply(self):
        """Implement this function add apply() logic after this class saves the current config."""


class YamlStringAction(SimpleConfigAction):

    def __init__(self, yaml_str):
        super().__init__()
        self._yaml_str = yaml_str

    def _apply(self):
        load_from_yaml(self._yaml_str)


class SetGeneralSampler(SimpleConfigAction):

    def __init__(self, sampler_name, settings=None):
        super().__init__()
        self._sampler_name = sampler_name
        self._settings = settings or {}

    def _apply(self):
        SampleSettingsLoader().load({'general': {'name': self._sampler_name,
                                                 'settings': self._settings}})


class SetGeneralOptimizer(SimpleConfigAction):

    def __init__(self, optimizer_name, settings=None):
        super().__init__()
        self._optimizer_name = optimizer_name
        self._settings = settings or {}

    @classmethod
    def from_object(self, optimizer):
        return SetGeneralOptimizer(optimizer.__class__.__name__, optimizer.optimizer_settings)

    def _apply(self):
        OptimizationSettingsLoader().load({'general': {'name': self._optimizer_name,
                                                       'settings': self._settings}})


"""Load the default configuration, and if possible, the users configuration."""
load_builtin()
try:
    load_user_home()
except IOError:
    pass



from mdt import CompositeModelTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Tensor(CompositeModelTemplate):

    description = 'The standard Tensor model with in vivo defaults.'
    model_expression = '''
        S0 * Tensor
    '''
    inits = {'Tensor.d': 1.7e-9,
             'Tensor.dperp0': 1.7e-10,
             'Tensor.dperp1': 1.7e-10}
    volume_selection = {'unweighted_threshold': 25e6,
                        'use_unweighted': True,
                        'use_weighted': True,
                        'min_bval': 0,
                        'max_bval': 1.5e9 + 0.1e9}


class NonParametricTensor(CompositeModelTemplate):

    description = 'The Tensor non-parameterized.'
    model_expression = '''
        S0 * SymmetricNonParametricTensor(Tensor)
    '''

    inits = {'Tensor.D_00': 0.3e-9, 'Tensor.D_11': 0.3e-9, 'Tensor.D_22': 0.3e-9,
             'Tensor.D_01': 0, 'Tensor.D_02': 0, 'Tensor.D_12': 0}

    lower_bounds = {'Tensor.D_00': 0, 'Tensor.D_11': 0, 'Tensor.D_22': 0,
                    'Tensor.D_01': -1e-9, 'Tensor.D_02': -1e-9, 'Tensor.D_12': -1e-9}

    upper_bounds = {'Tensor.D_00': 5e-9, 'Tensor.D_11': 5e-9, 'Tensor.D_22': 5e-9,
                    'Tensor.D_01': 1e-9, 'Tensor.D_02': 1e-9, 'Tensor.D_12': 1e-9}

    volume_selection = {'unweighted_threshold': 25e6,
                        'use_unweighted': True,
                        'use_weighted': True,
                        'min_bval': 0,
                        'max_bval': 1.5e9 + 0.1e9}



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'MapSpecificOptions.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MapSpecificOptions(object):
    def setupUi(self, MapSpecificOptions):
        MapSpecificOptions.setObjectName("MapSpecificOptions")
        MapSpecificOptions.resize(648, 590)
        self.gridLayout = QtWidgets.QGridLayout(MapSpecificOptions)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setHorizontalSpacing(0)
        self.gridLayout.setVerticalSpacing(10)
        self.gridLayout.setObjectName("gridLayout")
        self.frame_3 = CollapsablePanel(MapSpecificOptions)
        self.frame_3.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_3.setObjectName("frame_3")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.frame_3)
        self.verticalLayout_7.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_7.setSpacing(0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.label_22 = CollapsablePanelHeader(self.frame_3)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_22.setFont(font)
        self.label_22.setObjectName("label_22")
        self.verticalLayout_7.addWidget(self.label_22)
        self.line_5 = QtWidgets.QFrame(self.frame_3)
        self.line_5.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_5.setObjectName("line_5")
        self.verticalLayout_7.addWidget(self.line_5)
        self.frame_6 = CollapsablePanelContent(self.frame_3)
        self.frame_6.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_6.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_6.setObjectName("frame_6")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.frame_6)
        self.gridLayout_6.setContentsMargins(6, 6, 0, 0)
        self.gridLayout_6.setHorizontalSpacing(6)
        self.gridLayout_6.setVerticalSpacing(3)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.gridLayout_7 = QtWidgets.QGridLayout()
        self.gridLayout_7.setContentsMargins(-1, 0, -1, 0)
        self.gridLayout_7.setObjectName("gridLayout_7")
        self.label = QtWidgets.QLabel(self.frame_6)
        self.label.setObjectName("label")
        self.gridLayout_7.addWidget(self.label, 0, 0, 1, 1)
        self.use_data_scale_min = QtWidgets.QCheckBox(self.frame_6)
        self.use_data_scale_min.setText("")
        self.use_data_scale_min.setObjectName("use_data_scale_min")
        self.gridLayout_7.addWidget(self.use_data_scale_min, 1, 2, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.frame_6)
        self.label_5.setObjectName("label_5")
        self.gridLayout_7.addWidget(self.label_5, 1, 0, 1, 1)
        self.use_data_scale_max = QtWidgets.QCheckBox(self.frame_6)
        self.use_data_scale_max.setText("")
        self.use_data_scale_max.setObjectName("use_data_scale_max")
        self.gridLayout_7.addWidget(self.use_data_scale_max, 0, 2, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.frame_6)
        self.label_4.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_4.setObjectName("label_4")
        self.gridLayout_7.addWidget(self.label_4, 2, 1, 1, 1)
        self.data_set_use_scale = QtWidgets.QCheckBox(self.frame_6)
        self.data_set_use_scale.setText("")
        self.data_set_use_scale.setObjectName("data_set_use_scale")
        self.gridLayout_7.addWidget(self.data_set_use_scale, 2, 2, 1, 1)
        self.data_scale_max = ScientificNumberScroller(self.frame_6)
        self.data_scale_max.setObjectName("data_scale_max")
        self.gridLayout_7.addWidget(self.data_scale_max, 0, 1, 1, 1)
        self.data_scale_min = ScientificNumberScroller(self.frame_6)
        self.data_scale_min.setObjectName("data_scale_min")
        self.gridLayout_7.addWidget(self.data_scale_min, 1, 1, 1, 1)
        self.gridLayout_6.addLayout(self.gridLayout_7, 0, 0, 1, 1)
        self.verticalLayout_7.addWidget(self.frame_6)
        self.gridLayout.addWidget(self.frame_3, 1, 0, 1, 1)
        self.frame = CollapsablePanel(MapSpecificOptions)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame.setObjectName("frame")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setSpacing(0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_18 = CollapsablePanelHeader(self.frame)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_18.setFont(font)
        self.label_18.setObjectName("label_18")
        self.verticalLayout_5.addWidget(self.label_18)
        self.line_3 = QtWidgets.QFrame(self.frame)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout_5.addWidget(self.line_3)
        self.frame_4 = CollapsablePanelContent(self.frame)
        self.frame_4.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_4.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_4.setObjectName("frame_4")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.frame_4)
        self.gridLayout_3.setContentsMargins(6, 6, 0, 0)
        self.gridLayout_3.setHorizontalSpacing(6)
        self.gridLayout_3.setVerticalSpacing(3)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label_3 = QtWidgets.QLabel(self.frame_4)
        self.label_3.setObjectName("label_3")
        self.gridLayout_3.addWidget(self.label_3, 2, 0, 1, 1)
        self.colormap = QtWidgets.QComboBox(self.frame_4)
        self.colormap.setObjectName("colormap")
        self.gridLayout_3.addWidget(self.colormap, 3, 1, 1, 1)
        self.label_9 = QtWidgets.QLabel(self.frame_4)
        self.label_9.setObjectName("label_9")
        self.gridLayout_3.addWidget(self.label_9, 3, 0, 1, 1)
        self.label_21 = QtWidgets.QLabel(self.frame_4)
        self.label_21.setObjectName("label_21")
        self.gridLayout_3.addWidget(self.label_21, 0, 0, 1, 1)
        self.map_title = QtWidgets.QPlainTextEdit(self.frame_4)
        self.map_title.setObjectName("map_title")
        self.gridLayout_3.addWidget(self.map_title, 0, 1, 1, 1)
        self.data_colorbar_label = QtWidgets.QPlainTextEdit(self.frame_4)
        self.data_colorbar_label.setObjectName("data_colorbar_label")
        self.gridLayout_3.addWidget(self.data_colorbar_label, 2, 1, 1, 1)
        self.verticalLayout_5.addWidget(self.frame_4)
        self.gridLayout.addWidget(self.frame, 0, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 4, 0, 1, 1)
        self.frame_2 = CollapsablePanel(MapSpecificOptions)
        self.frame_2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_2.setObjectName("frame_2")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.frame_2)
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_6.setSpacing(0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_20 = CollapsablePanelHeader(self.frame_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_20.setFont(font)
        self.label_20.setObjectName("label_20")
        self.verticalLayout_6.addWidget(self.label_20)
        self.line_4 = QtWidgets.QFrame(self.frame_2)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.verticalLayout_6.addWidget(self.line_4)
        self.frame_5 = CollapsablePanelContent(self.frame_2)
        self.frame_5.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_5.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_5.setObjectName("frame_5")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.frame_5)
        self.gridLayout_4.setContentsMargins(6, 6, 0, 0)
        self.gridLayout_4.setHorizontalSpacing(6)
        self.gridLayout_4.setVerticalSpacing(3)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.info_file_location = QtWidgets.QLabel(self.frame_5)
        self.info_file_location.setWordWrap(True)
        self.info_file_location.setObjectName("info_file_location")
        self.gridLayout_4.addWidget(self.info_file_location, 0, 1, 1, 1)
        self.label_11 = QtWidgets.QLabel(self.frame_5)
        self.label_11.setObjectName("label_11")
        self.gridLayout_4.addWidget(self.label_11, 3, 0, 1, 1)
        self.label_24 = QtWidgets.QLabel(self.frame_5)
        self.label_24.setObjectName("label_24")
        self.gridLayout_4.addWidget(self.label_24, 0, 0, 1, 1)
        self.info_minimum = QtWidgets.QLabel(self.frame_5)
        self.info_minimum.setObjectName("info_minimum")
        self.gridLayout_4.addWidget(self.info_minimum, 3, 1, 1, 1)
        self.label_6 = QtWidgets.QLabel(self.frame_5)
        self.label_6.setObjectName("label_6")
        self.gridLayout_4.addWidget(self.label_6, 1, 0, 1, 1)
        self.info_shape = QtWidgets.QLabel(self.frame_5)
        self.info_shape.setObjectName("info_shape")
        self.gridLayout_4.addWidget(self.info_shape, 1, 1, 1, 1)
        self.label_27 = QtWidgets.QLabel(self.frame_5)
        self.label_27.setObjectName("label_27")
        self.gridLayout_4.addWidget(self.label_27, 5, 0, 1, 1)
        self.info_has_nan = QtWidgets.QLabel(self.frame_5)
        self.info_has_nan.setObjectName("info_has_nan")
        self.gridLayout_4.addWidget(self.info_has_nan, 5, 1, 1, 1)
        self.label_26 = QtWidgets.QLabel(self.frame_5)
        self.label_26.setObjectName("label_26")
        self.gridLayout_4.addWidget(self.label_26, 2, 0, 1, 1)
        self.info_maximum = QtWidgets.QLabel(self.frame_5)
        self.info_maximum.setObjectName("info_maximum")
        self.gridLayout_4.addWidget(self.info_maximum, 2, 1, 1, 1)
        self.gridLayout_4.setColumnStretch(1, 1)
        self.verticalLayout_6.addWidget(self.frame_5)
        self.gridLayout.addWidget(self.frame_2, 3, 0, 1, 1)
        self.info_Clipping = CollapsablePanel(MapSpecificOptions)
        self.info_Clipping.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.info_Clipping.setFrameShadow(QtWidgets.QFrame.Plain)
        self.info_Clipping.setObjectName("info_Clipping")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.info_Clipping)
        self.verticalLayout_8.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_8.setSpacing(0)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.label_25 = CollapsablePanelHeader(self.info_Clipping)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_25.setFont(font)
        self.label_25.setObjectName("label_25")
        self.verticalLayout_8.addWidget(self.label_25)
        self.line_6 = QtWidgets.QFrame(self.info_Clipping)
        self.line_6.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_6.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_6.setObjectName("line_6")
        self.verticalLayout_8.addWidget(self.line_6)
        self.frame_8 = CollapsablePanelContent(self.info_Clipping)
        self.frame_8.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_8.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_8.setObjectName("frame_8")
        self.gridLayout_8 = QtWidgets.QGridLayout(self.frame_8)
        self.gridLayout_8.setContentsMargins(6, 6, 0, 0)
        self.gridLayout_8.setHorizontalSpacing(6)
        self.gridLayout_8.setVerticalSpacing(3)
        self.gridLayout_8.setObjectName("gridLayout_8")
        self.gridLayout_9 = QtWidgets.QGridLayout()
        self.gridLayout_9.setContentsMargins(-1, 0, -1, 0)
        self.gridLayout_9.setObjectName("gridLayout_9")
        self.data_clipping_min = ScientificNumberScroller(self.frame_8)
        self.data_clipping_min.setObjectName("data_clipping_min")
        self.gridLayout_9.addWidget(self.data_clipping_min, 1, 1, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.frame_8)
        self.label_7.setObjectName("label_7")
        self.gridLayout_9.addWidget(self.label_7, 0, 0, 1, 1)
        self.data_set_use_clipping = QtWidgets.QCheckBox(self.frame_8)
        self.data_set_use_clipping.setText("")
        self.data_set_use_clipping.setObjectName("data_set_use_clipping")
        self.gridLayout_9.addWidget(self.data_set_use_clipping, 2, 2, 1, 1)
        self.use_data_clipping_max = QtWidgets.QCheckBox(self.frame_8)
        self.use_data_clipping_max.setText("")
        self.use_data_clipping_max.setObjectName("use_data_clipping_max")
        self.gridLayout_9.addWidget(self.use_data_clipping_max, 0, 2, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.frame_8)
        self.label_2.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_2.setObjectName("label_2")
        self.gridLayout_9.addWidget(self.label_2, 2, 1, 1, 1)
        self.use_data_clipping_min = QtWidgets.QCheckBox(self.frame_8)
        self.use_data_clipping_min.setText("")
        self.use_data_clipping_min.setObjectName("use_data_clipping_min")
        self.gridLayout_9.addWidget(self.use_data_clipping_min, 1, 2, 1, 1)
        self.data_clipping_max = ScientificNumberScroller(self.frame_8)
        self.data_clipping_max.setObjectName("data_clipping_max")
        self.gridLayout_9.addWidget(self.data_clipping_max, 0, 1, 1, 1)
        self.label_8 = QtWidgets.QLabel(self.frame_8)
        self.label_8.setObjectName("label_8")
        self.gridLayout_9.addWidget(self.label_8, 1, 0, 1, 1)
        self.gridLayout_8.addLayout(self.gridLayout_9, 0, 0, 1, 1)
        self.verticalLayout_8.addWidget(self.frame_8)
        self.gridLayout.addWidget(self.info_Clipping, 2, 0, 1, 1)

        self.retranslateUi(MapSpecificOptions)
        QtCore.QMetaObject.connectSlotsByName(MapSpecificOptions)
        MapSpecificOptions.setTabOrder(self.colormap, self.use_data_scale_max)
        MapSpecificOptions.setTabOrder(self.use_data_scale_max, self.use_data_scale_min)
        MapSpecificOptions.setTabOrder(self.use_data_scale_min, self.data_set_use_scale)
        MapSpecificOptions.setTabOrder(self.data_set_use_scale, self.use_data_clipping_max)
        MapSpecificOptions.setTabOrder(self.use_data_clipping_max, self.use_data_clipping_min)
        MapSpecificOptions.setTabOrder(self.use_data_clipping_min, self.data_set_use_clipping)

    def retranslateUi(self, MapSpecificOptions):
        _translate = QtCore.QCoreApplication.translate
        MapSpecificOptions.setWindowTitle(_translate("MapSpecificOptions", "Form"))
        self.label_22.setText(_translate("MapSpecificOptions", "Scale"))
        self.label.setText(_translate("MapSpecificOptions", "Max:"))
        self.label_5.setText(_translate("MapSpecificOptions", "Min:"))
        self.label_4.setText(_translate("MapSpecificOptions", "Enable:"))
        self.label_18.setText(_translate("MapSpecificOptions", "General"))
        self.label_3.setText(_translate("MapSpecificOptions", "Colorbar label:"))
        self.label_9.setText(_translate("MapSpecificOptions", "Colormap:"))
        self.label_21.setText(_translate("MapSpecificOptions", "Title (latex):"))
        self.label_20.setText(_translate("MapSpecificOptions", "Info"))
        self.info_file_location.setText(_translate("MapSpecificOptions", "TextLabel"))
        self.label_11.setText(_translate("MapSpecificOptions", "Minimum:"))
        self.label_24.setText(_translate("MapSpecificOptions", "Filename:"))
        self.info_minimum.setText(_translate("MapSpecificOptions", "TextLabel"))
        self.label_6.setText(_translate("MapSpecificOptions", "Shape:"))
        self.info_shape.setText(_translate("MapSpecificOptions", "TextLabel"))
        self.label_27.setText(_translate("MapSpecificOptions", "Has NaNs:"))
        self.info_has_nan.setText(_translate("MapSpecificOptions", "TextLabel"))
        self.label_26.setText(_translate("MapSpecificOptions", "Maximum:"))
        self.info_maximum.setText(_translate("MapSpecificOptions", "TextLabel"))
        self.label_25.setText(_translate("MapSpecificOptions", "Clipping"))
        self.label_7.setText(_translate("MapSpecificOptions", "Max:"))
        self.label_2.setText(_translate("MapSpecificOptions", "Enable:"))
        self.label_8.setText(_translate("MapSpecificOptions", "Min:"))

from ..widgets import CollapsablePanel, CollapsablePanelContent, CollapsablePanelHeader
from mdt.gui.widgets.composite_widgets import ScientificNumberScroller



from mdt import CascadeTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BallRacket_r1_c(CascadeTemplate):

    description = 'Initializing the Ball&Rackets using Ball&Stick.'
    models = ('BallStick_r1 (Cascade)',
              'BallRacket_r1')
    inits = {'BallRacket_r1': {'Racket0.theta': 'Stick0.theta',
                               'Racket0.phi': 'Stick0.phi',
                               'w_res0.w': 'w_stick0.w'}}


class BallRacket_r2(CascadeTemplate):

    description = 'Cascade for BallRacket_r2.'
    models = ('BallStick_r2 (Cascade)',
              'BallRacket_r1 (Cascade)',
              'BallRacket_r2')
    inits = {'BallRacket_r2': {
        'w_res1.w': 0.0,
        'Racket1.theta': lambda _, out_all: out_all[0]['Stick1.theta'],
        'Racket1.phi': lambda _, out_all: out_all[0]['Stick1.phi']}}


class BallRacket_r2_fixed(CascadeTemplate):

    description = 'Cascade for BallRacket_r2.'
    cascade_name_modifier = 'fixed'
    models = ('BallStick_r2 (Cascade)',
              'BallRacket_r1 (Cascade)',
              'BallRacket_r2')
    inits = {'BallRacket_r2': {'w_res1.w': 0.0}}
    fixes = {'BallRacket_r2': {
        'Racket0.theta': lambda _, out_all: out_all[0]['Stick0.theta'],
        'Racket0.phi': lambda _, out_all: out_all[0]['Stick0.phi'],
        'Racket1.theta': lambda _, out_all: out_all[0]['Stick1.theta'],
        'Racket1.phi': lambda _, out_all: out_all[0]['Stick1.phi']
    }}



from mdt import BatchProfileTemplate

__author__ = 'Robbert Harms'
__date__ = '2018-03-18'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class HCP_MGH(BatchProfileTemplate):
    description = '''
        The profile for the MGH data from the Human Connectome project.

        This assumes that you downloaded and extracted the MGH data in one folder and that you now have one folder per subject.

        You can provide the noise standard deviation to use using a noise_std file containing a single float.

        Example directory layout:
            /mgh_*/diff/preproc/mri/diff_preproc.nii(.gz)
            /mgh_*/diff/preproc/bvals.txt
            /mgh_*/diff/preproc/bvecs_fsl_moco_norm.txt

        Optional items:
            /mgh_*/diff/preproc/diff_preproc.prtcl
            /mgh_*/diff/preproc/mri/diff_preproc_mask.nii(.gz)
            /mgh_*/diff/preproc/noise_std.{txt,nii,nii.gz}
    '''
    subject_base_folder = '{subject_id}/diff/preproc/'

    data_fname = '{subject_base_folder}/mri/diff_preproc.nii*'
    mask_fname = '{subject_base_folder}/mri/diff_preproc_mask.nii*'

    bvec_fname = '{subject_base_folder}/bvecs_fsl_moco_norm.txt'
    bval_fname = '{subject_base_folder}/bvals.txt'
    protocol_columns = {'Delta': 21.8e-3, 'delta': 12.9e-3, 'TR': 8800e-3, 'TE': 57e-3}


class HCP_WUMINN(BatchProfileTemplate):
    description = '''
        The profile for the WU-Minn data from the Human Connectome project',

        This assumes that you downloaded and extracted the WU-Minn data in one folder which gives one folder per subject.

        You can provide the noise standard deviation to use using a noise_std file containing a single float.

        Example directory layout:
            /*/T1w/Diffusion/data.nii.gz
            /*/T1w/Diffusion/bvals
            /*/T1w/Diffusion/bvecs
            /*/T1w/Diffusion/nodif_brain_mask.nii.gz
            /*/T1w/Diffusion/grad_dev.nii.gz

        Optional items:
            /*/T1w/Diffusion/data.prtcl
            /*/T1w/Diffusion/noise_std
    '''
    subject_base_folder = '{subject_id}/T1w/Diffusion/'

    data_fname = '{subject_base_folder}/data.nii*'
    mask_fname = '{subject_base_folder}/nodif_brain_mask.nii*'

    gradient_deviations_fname = '{subject_base_folder}/grad_dev.nii*'

    bvec_fname = '{subject_base_folder}/bvecs'
    bval_fname = '{subject_base_folder}/bvals'
    protocol_columns = {'Delta': 43.1e-3, 'delta': 10.6e-3, 'TE': 89.5e-3, 'TR': 5520e-3}



import logging
from importlib.machinery import SourceFileLoader
import inspect
import os
from collections import defaultdict
from contextlib import contextmanager
import mdt
import mot
from mdt.configuration import get_config_dir
from mot.library_functions import CLLibrary
from mdt.model_building.likelihood_functions import LikelihoodFunction
from mdt.model_building.parameter_functions.transformations import AbstractTransformation

__author__ = 'Robbert Harms'
__date__ = '2018-03-22'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


supported_component_types = ('batch_profiles', 'cascade_models', 'compartment_models',
                             'composite_models', 'library_functions', 'parameters', 'likelihood_functions',
                             'parameter_transforms')


class _ComponentLibrary(object):

    def __init__(self):
        """Holds the reference to all defined components, by component type and by name.

        For each component type several components may be defined with different or equal names. If the names are equal
        they are added to a stack and only the last element is returned. Components may also be removed again from
        the stack (in a random access method).
        """
        self._library = {}
        self._mutation_history = []
        self.reset()

    def get_current_history_length(self):
        """Get the current length of the history. Useful for undoing history changes."""
        return len(self._mutation_history)

    def undo_history_until(self, history_ind):
        """Pop the history stack until we are at the given index (length).

        Args:
            history_ind (int): the desired length of the history stack
        """
        if history_ind < len(self._mutation_history):
            for ind in range(len(self._mutation_history) - history_ind):
                item = self._mutation_history.pop()
                if item.action == 'add':
                    self._library[item.component_type][item.name].pop()
                if item.action == 'remove':
                    self._library[item.component_type][item.name].append(item.adapter)

    def reset(self):
        """Clear the library by removing all available components.
        This also resets the mutation history.
        """
        self._library = {component_type: defaultdict(list) for component_type in supported_component_types}
        self._mutation_history = []

    def add_component(self, component_type, name, component_class, meta_info=None):
        """Adds a component class to the library.

        Args:
            component_type (str): the type of the component, see ``supported_component_types``.
            name (str): the name of the component
            component_class (class): the class or constructor function for the component
            meta_info (dict): a dictionary with meta information about the component
        """
        adapter = _DirectComponent(component_class, meta_info=meta_info)
        self._library[component_type][name].append(adapter)
        self._mutation_history.append(_LibraryHistoryDelta('add', component_type, name, adapter))

    def add_template_component(self, template):
        """Adds a component template to the library.

        Args:
            template (mdt.component_templates.base.ComponentTemplateMeta): the template for constructing the component
                class.
        """
        adapter = _ComponentFromTemplate(template)
        self._library[template.component_type][template.name].append(adapter)
        self._mutation_history.append(_LibraryHistoryDelta('add', template.component_type, template.name, adapter))

    def get_component(self, component_type, name):
        """Get the component class for the component of the given type and name.

        Args:
            component_type (str): the type of the component, see ``supported_component_types``.
            name (str): the name of the component

        Returns:
            class: the component class.
        """
        if not self.has_component(component_type, name):
            raise ValueError('Can not find a component of type "{}" with name "{}"'.format(component_type, name))
        return self._library[component_type][name][-1].get_component()

    def get_meta_info(self, component_type, name):
        """Get the meta information dictionary for the component of the given type and name.

        Args:
            component_type (str): the type of the component, see ``supported_component_types``.
            name (str): the name of the component

        Returns:
            dict: the meta information
        """
        if not self.has_component(component_type, name):
            raise ValueError('Can not find a component of type "{}" with name "{}"'.format(component_type, name))
        return self._library[component_type][name][-1].get_meta_info()

    def get_component_list(self, component_type):
        """Get a list of available components by component type.

        Args:
            component_type (str): the type of the component, see ``supported_component_types``.

        Returns:
            list of str: list of available components
        """
        return list(self._library[component_type].keys())

    def has_component(self, component_type, name):
        """Check if a component is available.

        Args:
            component_type (str): the type of the component, see ``supported_component_types``.
            name (str): the name of the component

        Returns:
            boolean: if we have a component available of the given type and given name.
        """
        return name in self._library[component_type] and len(self._library[component_type][name])

    def get_template(self, component_type, name):
        """Get the template class for the given component.

        This may not be supported for all component types and components. That is, since components can either be
        added as classes or as templates, we can not guarantee a template class for any requested component.

        Args:
            component_type (str): the type of the component, see ``supported_component_types``.
            name (str): the name of the component

        Returns:
            mdt.component_templates.base.ComponentTemplateMeta: a template class if possible.

        Raises:
            ValueError: if no component of the given name could be found.
        """
        if not self.has_component(component_type, name):
            raise ValueError('The component with the name "{}" '
                             'of type "{}" could be found.'.format(name, component_type))
        return self._library[component_type][name][-1].get_template()

    def remove_last_entry(self, component_type, name):
        """Removes the last entry of the given component.

        Args:
            component_type (str): the type of the component, see ``supported_component_types``.
            name (str): the name of the component
        """
        adapter = self._library[component_type][name].pop()
        if not len(self._library[component_type][name]):
            del self._library[component_type][name]
        self._mutation_history.append(_LibraryHistoryDelta('remove', component_type, name, adapter))


class _ComponentAdapter(object):

    def get_component(self):
        """Build or return the actual component class.

        Since the component library supports both ``component classes`` as ``template classes`` we need an adapter class
        to build the actual component if necessary.

        Returns:
            class: the component class
        """
        raise NotImplementedError()

    def get_meta_info(self):
        """Get the meta info of this component

        Returns:
            dict: the meta info
        """
        raise NotImplementedError()

    def get_template(self):
        """If supported, gets the template of this component.

        Returns:
            mdt.component_templates.base.ComponentTemplateMeta: a template class if possible.
        """
        raise NotImplementedError()


class _DirectComponent(_ComponentAdapter):

    def __init__(self, component, meta_info=None):
        self.component = component
        self.meta_info = meta_info or {}

    def get_component(self):
        return self.component

    def get_meta_info(self):
        return self.meta_info

    def get_template(self):
        raise ValueError('Can not build template from component class.')


class _ComponentFromTemplate(_ComponentAdapter):

    def __init__(self, template):
        self.template = template

    def get_component(self):
        return self.template()

    def get_meta_info(self):
        return self.template.meta_info()

    def get_template(self):
        return self.template


class _LibraryHistoryDelta(object):

    def __init__(self, action, component_type, name, adapter):
        """Representation of a history change in the component library.

        Args:
            action (str): one of ``remove`` or ``add``.
            component_type (str): the type of the component
            name (str): the name of the component.
            adapter (_ComponentAdapter): the adapter instance
        """
        self.component_type = component_type
        self.name = name
        self.adapter = adapter
        self.action = action


component_library = _ComponentLibrary()


def _add_doc(value):
    """Add a docstring to the given value."""
    def _doc(func):
        func.__doc__ = value
        return func
    return _doc


@_add_doc(_ComponentLibrary.add_component.__doc__)
def add_component(component_type, name, cls, meta_info=None):
    return component_library.add_component(component_type, name, cls, meta_info)


@_add_doc(_ComponentLibrary.add_template_component.__doc__)
def add_template_component(template):
    return component_library.add_template_component(template)


@_add_doc(_ComponentLibrary.get_template.__doc__)
def get_template(component_type, name):
    return component_library.get_template(component_type, name)


@_add_doc(_ComponentLibrary.get_component.__doc__)
def get_component(component_type, name):
    return component_library.get_component(component_type, name)


@_add_doc(_ComponentLibrary.has_component.__doc__)
def has_component(component_type, name):
    return component_library.has_component(component_type, name)


@_add_doc(_ComponentLibrary.get_component_list.__doc__)
def get_component_list(component_type):
    return component_library.get_component_list(component_type)


@_add_doc(_ComponentLibrary.get_meta_info.__doc__)
def get_meta_info(component_type, name):
    return component_library.get_meta_info(component_type, name)


@_add_doc(_ComponentLibrary.remove_last_entry.__doc__)
def remove_last_entry(component_type, name):
    return component_library.remove_last_entry(component_type, name)


@contextmanager
def temporary_component_updates():
    """Creates a context that keeps track of the component mutations and undoes them when the context exits.

    This can be useful to temporarily add or remove some components from the library.
    """
    history_ind = component_library.get_current_history_length()
    yield
    component_library.undo_history_until(history_ind)


def reload():
    """Clear the component library and reload all default components.

    This will load the components from the user home folder and from the MOT library.
    """
    component_library.reset()
    _load_mot_components()
    _load_home_folder()
    _load_automatic_cascades()


def get_model(model_name):
    """Load the class of one of the available models.

    Args:
        model_name (str): One of the models from the cascade models or composite models

    Returns:
        class: Either a cascade model or a composite model. In any case, a model that can be given to the ``fit_model``
            function.
    """
    try:
        return component_library.get_component('cascade_models', model_name)
    except ValueError:
        try:
            return component_library.get_component('composite_models', model_name)
        except ValueError:
            raise ValueError('The model with the name "{}" could not be found.'.format(model_name))


def list_composite_models():
    """Get a name listing of all available composite models.

    Returns:
        list of str: a list of available composite model names
    """
    return component_library.get_component_list('composite_models')


def list_cascade_models(target_model_name=None):
    """Get a list of all available cascade models

    Args:
        target_model_name (str): if given we will only return the list of cascades that end with this composite model.

    Returns:
        list of str: A list of available cascade models
    """
    model_names = component_library.get_component_list('cascade_models')

    if target_model_name:
        cascades = []
        for name in model_names:
            meta_info = get_meta_info('cascade_models', name)
            if meta_info['target_model'] == target_model_name:
                cascades.append(name)
        return cascades

    return model_names


def get_batch_profile(batch_profile):
    """Load the class of one of the available batch profiles

    Args:
        batch_profile (str): The name of the batch profile class to load

    Returns:
        cls: the batch profile class
    """
    return component_library.get_component('batch_profiles', batch_profile)


def _load_mot_components():
    """Load all the components from MOT."""
    items = [
        (mot.library_functions, CLLibrary, 'library_functions'),
        (mdt.model_building.likelihood_functions, LikelihoodFunction, 'likelihood_functions'),
        (mdt.model_building.parameter_functions.transformations, AbstractTransformation, 'parameter_transforms'),
    ]

    for module_obj, class_type, component_type in items:
        module_items = inspect.getmembers(module_obj, lambda cls: inspect.isclass(cls) and issubclass(cls, class_type))
        for item in [x[0] for x in module_items if x[0] != class_type.__name__]:
            add_component(component_type, item, getattr(module_obj, item))


def _load_home_folder():
    """Load the components from the MDT home folder.

    This first loads all components from the ``standard`` folder and next all those from the ``user`` folder.
    """
    for user_type in ['standard', 'user']:
        base_path = os.path.join(get_config_dir(), 'components', user_type)
        for path, sub_dirs, files in os.walk(base_path):
            for file in files:
                if file.endswith('.py') and not file.startswith('__'):
                    full_path = os.path.join(path, file)

                    module_name = os.path.splitext(full_path[len(os.path.join(get_config_dir(), 'components')):])[0]

                    try:
                        SourceFileLoader(module_name, full_path).load_module()
                    except Exception as e:
                        logger = logging.getLogger(__name__)
                        logger.warning('Could not load the file "{}", exception: "{}".'.format(full_path, str(e)))


def _load_automatic_cascades():
    """Automatically create cascade models where possible.

    This generates cascade models matching the scheme in Harms 2017: CS, CI and CF cascades:

    - CS: Cascade S0, a cascade which only initializes the model with an S0 estimate
    - CI: Cascade Initialized: initalizes the volume fractions and orientations
    - CF: Cascade Fixed: initializes the volume fractions, fixes the orientations
    """
    from mdt.configuration import use_automatic_generated_cascades, get_automatic_generated_cascades_excluded

    def get_missing_s0_cascades(models, cascades):
        missing_cascades = []
        for model_name in models:
            if '{} (Cascade|S0)'.format(model_name) not in cascades:
                missing_cascades.append('{} (Cascade|S0)'.format(model_name))
        return missing_cascades

    def generate_cascade(cascaded_name):
        from mdt.component_templates.cascade_models import CascadeTemplate

        if '(Cascade|S0)' in cascaded_name:
            class Template(CascadeTemplate):
                cascade_name_modifier = 'S0'
                description = 'Automatically generated cascade.'
                models = ('S0',
                          cascaded_name[0:-len('(Cascade|S0)')].strip())

    if use_automatic_generated_cascades():
        excludes = get_automatic_generated_cascades_excluded()
        models_list = [m for m in list_composite_models() if m not in excludes]

        for cascade in get_missing_s0_cascades(models_list, list_cascade_models()):
            generate_cascade(cascade)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'generate_protocol_load_gb_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_LoadGBDialog(object):
    def setupUi(self, LoadGBDialog):
        LoadGBDialog.setObjectName("LoadGBDialog")
        LoadGBDialog.resize(831, 227)
        self.verticalLayout = QtWidgets.QVBoxLayout(LoadGBDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(LoadGBDialog)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(LoadGBDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_3.addWidget(self.label_4)
        self.verticalLayout.addLayout(self.verticalLayout_3)
        self.line = QtWidgets.QFrame(LoadGBDialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.bvalFileChooser = QtWidgets.QPushButton(LoadGBDialog)
        self.bvalFileChooser.setEnabled(True)
        self.bvalFileChooser.setObjectName("bvalFileChooser")
        self.horizontalLayout.addWidget(self.bvalFileChooser)
        self.bvalFileInput = QtWidgets.QLineEdit(LoadGBDialog)
        self.bvalFileInput.setEnabled(True)
        self.bvalFileInput.setObjectName("bvalFileInput")
        self.horizontalLayout.addWidget(self.bvalFileInput)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 1, 1, 1)
        self.label_11 = QtWidgets.QLabel(LoadGBDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_11.setFont(font)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 0, 2, 1, 1)
        self.label_5 = QtWidgets.QLabel(LoadGBDialog)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 0, 0, 1, 1)
        self.label_6 = QtWidgets.QLabel(LoadGBDialog)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 1, 0, 1, 1)
        self.label_12 = QtWidgets.QLabel(LoadGBDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_12.setFont(font)
        self.label_12.setObjectName("label_12")
        self.gridLayout.addWidget(self.label_12, 1, 2, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.bvecFileChooser = QtWidgets.QPushButton(LoadGBDialog)
        self.bvecFileChooser.setObjectName("bvecFileChooser")
        self.horizontalLayout_2.addWidget(self.bvecFileChooser)
        self.bvecFileInput = QtWidgets.QLineEdit(LoadGBDialog)
        self.bvecFileInput.setObjectName("bvecFileInput")
        self.horizontalLayout_2.addWidget(self.bvecFileInput)
        self.gridLayout.addLayout(self.horizontalLayout_2, 0, 1, 1, 1)
        self.label_13 = QtWidgets.QLabel(LoadGBDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_13.setFont(font)
        self.label_13.setObjectName("label_13")
        self.gridLayout.addWidget(self.label_13, 2, 2, 1, 1)
        self.bvalScale = QtWidgets.QLineEdit(LoadGBDialog)
        self.bvalScale.setObjectName("bvalScale")
        self.gridLayout.addWidget(self.bvalScale, 2, 1, 1, 1)
        self.label_7 = QtWidgets.QLabel(LoadGBDialog)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 2, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.line_3 = QtWidgets.QFrame(LoadGBDialog)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.buttonBox = QtWidgets.QDialogButtonBox(LoadGBDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(LoadGBDialog)
        self.buttonBox.accepted.connect(LoadGBDialog.accept)
        self.buttonBox.rejected.connect(LoadGBDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(LoadGBDialog)
        LoadGBDialog.setTabOrder(self.bvalFileChooser, self.bvalFileInput)

    def retranslateUi(self, LoadGBDialog):
        _translate = QtCore.QCoreApplication.translate
        LoadGBDialog.setWindowTitle(_translate("LoadGBDialog", "Load g & b"))
        self.label_3.setText(_translate("LoadGBDialog", "Load g & b"))
        self.label_4.setText(_translate("LoadGBDialog", "Load the bvec (g) and bval (b) in the protocol"))
        self.bvalFileChooser.setText(_translate("LoadGBDialog", "Browse"))
        self.label_11.setText(_translate("LoadGBDialog", "(The file containing the gradient directions)"))
        self.label_5.setText(_translate("LoadGBDialog", "Bvec (g) file:"))
        self.label_6.setText(_translate("LoadGBDialog", "Bval (b) file:"))
        self.label_12.setText(_translate("LoadGBDialog", "(The file containing the b-values)"))
        self.bvecFileChooser.setText(_translate("LoadGBDialog", "Browse"))
        self.label_13.setText(_translate("LoadGBDialog", "(We expect the b-values in the\n"
"protocol in units of s/m^2)"))
        self.bvalScale.setText(_translate("LoadGBDialog", "1e6"))
        self.label_7.setText(_translate("LoadGBDialog", "B-value rescale:"))

